import os


def load_env_variables(silent=False):
    try:
        # FIXME dotenv is not working, there is some dotenv plugin for poetry -> solve it more elegantly
        with open(".env") as input_file:
            for line in input_file:
                if not line.strip():  # ignoring empty lines
                    continue
                if line.startswith("#"):  # ignoring comment lines
                    continue
                key, value = line.partition("=")[::2]
                os.environ[key.strip()] = str(value).strip()
                if not silent:
                    print(
                        f"added env variable: {key.strip()}={str(value).strip()} (env reread: {os.environ.get(key.strip())})"
                    )
    except FileNotFoundError:
        print("input env file not present")




    
from pydantic import BaseModel, Field, IPvAnyAddress
from typing import List, Optional, Dict, Union, Literal
from vrouter_agent.utils.config import get_frr_sync_dir
from ipaddress import IPv4Address, IPv4Network
from .frr_config import InitialFRRConfig, NodeConfig


class Multichain(BaseModel):
    """
    Multichain config defined in the YAML file. 
    This will be used to determine the chain name and directory for the multichain node.
    """
    chain: str
    chain_dir: str
    stream: str
  

class CreatedWireguardTunnel(BaseModel):
    """
    Created Wireguard Tunnel model.
    """
    name: str
    ip_address: str
    peer_ip_address: Optional[str] = None
    mapped_name: Optional[str] = None
    
class Interface(BaseModel):
    """
    Interface config defined in the YAML file.
    """
    interface_name: str
    type: Literal['wan','lan']
    ip_address: Optional[IPvAnyAddress] = None
    prefix_len: Optional[int]
    subnet_mask: Optional[str]
    is_primary: bool
    gateway: Optional[str]
    mtu: Optional[int] = 1500


class GlobalSettings(BaseModel):
    environment: str = Field(default="development")
    user: str = Field(default="usdn-admin")
    secret_key: str = Field(default="usdn-admin") # to decrypt the data


class AppConfig(BaseModel):
    global_:  GlobalSettings = Field(..., alias="global")
    interfaces: List[Interface] = Field(default=[])
    multichain: Multichain = Field(default=Multichain(chain="multichain", chain_dir="/opt/multichain", stream="vrouter"))
    frr: InitialFRRConfig
    
    




def stream_receiver(stream, txid, *args, **kwargs):
    """
    Receives data from a stream and processes it using the provided callback function.
    
    :param stream: The stream to read data from.
    :param callback: The function to call with the received data.
    :param args: Additional positional arguments to pass to the callback.
    :param kwargs: Additional keyword arguments to pass to the callback.
    """
    from vrouter_agent.stream_handler import StreamHandler

    handler = StreamHandler(stream, txid)
    handler.start()

    try:
        while True:
            data = handler.read_data()
            if data is None:
                break
            handler.process_data(data, *args, **kwargs)
    except Exception as e:
        handler.handle_error(e)
    finally:
        handler.cleanup()
#     logger.info("Stream receiver started")
    


from logging.config import fileConfig
import sys
from pathlib import Path

from sqlalchemy import engine_from_config
from sqlalchemy import pool

from alembic import context

# Add the parent directory to sys.path to import our models
sys.path.append(str(Path(__file__).parent.parent.parent))

# Import SQLModel metadata from our models
from sqlmodel import SQLModel
from vrouter_agent.models.tunnel_config import TunnelConfigData, TunnelConfigHistory
from vrouter_agent.models.orders import Order
from vrouter_agent.models.devices import Device
from vrouter_agent.models.tunnels import Tunnel
from vrouter_agent.models.transactions import Transaction

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
target_metadata = SQLModel.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    # Use our existing database configuration
    from vrouter_agent.core.db import engine
    
    with engine.connect() as connection:
        context.configure(
            connection=connection, 
            target_metadata=target_metadata,
            # Enable batch mode for SQLite ALTER TABLE support
            render_as_batch=True
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



from fastapi import Depends, Form, BackgroundTasks
from loguru import logger
from sqlmodel import Session
from vrouter_agent.agent import VRouterAgent
from vrouter_agent.core.db import get_session
from vrouter_agent.services.chain import Chain
from vrouter_agent.models import Transaction
from vrouter_agent.core.config import settings 
from fastapi import APIRouter
from vrouter_agent.enhanced_stream_processor import get_stream_processor
from vrouter_agent.services.stream_listener import (
    get_stream_listener, 
    start_stream_listener, 
    stop_stream_listener,
    StreamListenerConfig
)
import time

router = APIRouter()

# Legacy route maintained for backwards compatibility
@router.post(
    "/",
)
async def transaction(
    background_tasks: BackgroundTasks,
    txid: str = Form(...),
    stream: str = Form(...),
    session: Session = Depends(get_session),
):
    """
    Legacy transaction endpoint maintained for backwards compatibility.
    
    Note: The system now primarily uses the stream listener service
    to automatically monitor and process transactions from the designated
    stream instead of relying on POST data.
    """
    logger.warning(f"Received transaction via legacy POST endpoint: {txid} on stream: {stream}")
    logger.info("Consider migrating to the stream listener service for automatic transaction processing")
    
    # Create transaction record
    new_transaction = Transaction(txid=txid, stream=stream)
    session.add(new_transaction)
    session.commit()
    logger.info(f"Transaction received: {new_transaction.id} on stream: {stream}")

    # Get the enhanced stream processor
    processor = await get_stream_processor()
        
    # Process transaction asynchronously using the enhanced processor
    success = await processor.process_transaction(new_transaction, stream, session)
    
    if success:
        logger.info(f"Transaction {new_transaction.id} queued for processing")
        return {
            "message": "Transaction queued for processing", 
            "id": new_transaction.id,
            "status": "queued"
        }
    else:
        logger.error(f"Failed to queue transaction {new_transaction.id}")
        return {
            "message": "Failed to queue transaction", 
            "id": new_transaction.id,
            "status": "error"
        }, 500

# Stream Listener Management Endpoints

# Endpoint to start the stream listener
@router.post("/stream-listener/start")
async def start_listener():
    """Start the stream listener service for monitoring the designated stream."""
    try:
        success = await start_stream_listener()
        
        if success:
            listener = await get_stream_listener()
            return {
                "message": "Stream listener started successfully",
                "stream_name": listener.config.stream_name,
                "hostname": listener.hostname,
                "status": "running"
            }
        else:
            return {
                "message": "Failed to start stream listener",
                "status": "error"
            }, 500
    except Exception as e:
        logger.error(f"Error starting stream listener: {e}")
        return {
            "message": f"Error starting stream listener: {str(e)}",
            "status": "error"
        }, 500

# Endpoint to stop the stream listener
@router.post("/stream-listener/stop")
async def stop_listener():
    """Stop the stream listener service."""
    try:
        await stop_stream_listener()
        return {
            "message": "Stream listener stopped successfully",
            "status": "stopped"
        }
    except Exception as e:
        logger.error(f"Error stopping stream listener: {e}")
        return {
            "message": f"Error stopping stream listener: {str(e)}",
            "status": "error"
        }, 500

# Endpoint to get stream listener status
@router.get("/stream-listener/status")
async def get_listener_status():
    """Get current status of the stream listener."""
    try:
        listener = await get_stream_listener()
        status = listener.get_status()
        metrics = listener.get_metrics()
        
        return {
            "status": status,
            "metrics": metrics,
            "timestamp": time.time()
        }
    except Exception as e:
        logger.error(f"Error getting stream listener status: {e}")
        return {
            "message": f"Error getting status: {str(e)}",
            "status": "error"
        }, 500

# Endpoint to get detailed processing statistics
@router.get("/stream-listener/stats")
async def get_listener_stats():
    """Get detailed processing statistics including database information."""
    try:
        listener = await get_stream_listener()
        stats = listener.get_processing_stats()
        
        return {
            "statistics": stats,
            "timestamp": time.time()
        }
    except Exception as e:
        logger.error(f"Error getting stream listener statistics: {e}")
        return {
            "message": f"Error getting statistics: {str(e)}",
            "status": "error"
        }, 500

# Endpoint to check if a specific transaction is processed
@router.get("/stream-listener/transaction/{txid}/processed")
async def check_transaction_processed(txid: str):
    """Check if a specific transaction has been processed."""
    try:
        listener = await get_stream_listener()
        is_processed = listener.is_transaction_processed(txid)
        
        return {
            "txid": txid,
            "processed": is_processed,
            "timestamp": time.time()
        }
    except Exception as e:
        logger.error(f"Error checking transaction {txid}: {e}")
        return {
            "message": f"Error checking transaction: {str(e)}",
            "txid": txid,
            "status": "error"
        }, 500

# Endpoint to manually mark a transaction as processed
@router.post("/stream-listener/transaction/{txid}/mark-processed")
async def mark_transaction_processed(txid: str):
    """Manually mark a transaction as processed."""
    try:
        listener = await get_stream_listener()
        listener.mark_transaction_processed(txid)
        
        return {
            "message": f"Transaction {txid} marked as processed",
            "txid": txid,
            "status": "success"
        }
    except Exception as e:
        logger.error(f"Error marking transaction {txid} as processed: {e}")
        return {
            "message": f"Error marking transaction: {str(e)}",
            "txid": txid,
            "status": "error"
        }, 500

# Endpoint to reload processed transactions from database
@router.post("/stream-listener/reload-cache")
async def reload_processed_cache():
    """Reload processed transactions cache from database."""
    try:
        listener = await get_stream_listener()
        listener.reload_processed_transactions()
        
        return {
            "message": "Processed transactions cache reloaded from database",
            "status": "success"
        }
    except Exception as e:
        logger.error(f"Error reloading cache: {e}")
        return {
            "message": f"Error reloading cache: {str(e)}",
            "status": "error"
        }, 500

# Endpoint to configure stream listener
@router.post("/stream-listener/configure")
async def configure_listener(
    stream_name: str = Form(None),
    poll_interval: float = Form(5.0),
    batch_size: int = Form(50),
    max_retries: int = Form(3)
):
    """Configure stream listener settings."""
    try:
        # Get current listener
        current_listener = await get_stream_listener()
        is_running = current_listener.is_running
        
        # Stop if running
        if is_running:
            await stop_stream_listener()
        
        # Create new configuration
        config = StreamListenerConfig(
            stream_name=stream_name or current_listener.hostname,
            poll_interval=poll_interval,
            batch_size=batch_size,
            max_retries=max_retries
        )
        
        # Start with new config
        success = await start_stream_listener(config)
        
        return {
            "message": "Stream listener configured successfully",
            "config": {
                "stream_name": config.stream_name,
                "poll_interval": config.poll_interval,
                "batch_size": config.batch_size,
                "max_retries": config.max_retries
            },
            "status": "running" if success else "error"
        }
    except Exception as e:
        logger.error(f"Error configuring stream listener: {e}")
        return {
            "message": f"Error configuring listener: {str(e)}",
            "status": "error"
        }, 500
# Legacy processing status endpoint
@router.get("/status")
async def get_processing_status():
    """Get current stream processor status and metrics."""
    processor = await get_stream_processor()
    status = processor.get_status()
    metrics = processor.get_metrics()
    
    return {
        "processor_status": status,
        "metrics": metrics,
        "timestamp": time.time()
    }

# Legacy transaction status endpoint
@router.get("/{transaction_id}/status")
async def get_transaction_status(transaction_id: int, session: Session = Depends(get_session)):
    """Get status of a specific transaction."""
    transaction = session.get(Transaction, transaction_id)
    if not transaction:
        return {"error": "Transaction not found"}, 404
    
    return {
        "transaction_id": transaction.id,
        "txid": transaction.txid,
        "stream": transaction.stream,
        "created_at": transaction.created_at,
        "status": "completed"  # This could be enhanced with actual status tracking
    }





from .cli import run_command


def check_service_status(name: str):
    """
    Checking vpp service status
    """
    try:
        run_command(["sudo", "systemctl", "is-active", "--quiet", name])
        return True
    except RuntimeError:
        return False


def restart_service(name: str):
    """
    Restarting vpp service
    """
    cmd = run_command(["sudo", "systemctl", "restart", name])
    return cmd



"""
Device and Interface models for FastAPI SQLModel.
"""

from typing import Optional, List
from sqlmodel import SQLModel, Field, Relationship
from enum import Enum
from .base import BaseModel


class NodeType(str, Enum):
    HUB = "hub"
    SPOKE = "spoke"
    PEER = "peer"
    EDGE = "edge"
    TRANSIT = "transit"
    ENDPOINT = "endpoint"


class DeviceStatus(str, Enum):
    OPERATIONAL = "operational"
    WARNING = "warning"
    CRITICAL = "critical"
    ERROR = "error"


class InterfaceType(str, Enum):
    PHYSICAL = "physical"
    VLAN = "vlan"
    LOOPBACK = "loopback"
    TUNNEL = "tunnel"
    BRIDGE = "bridge"


class Device(BaseModel, table=True):
    """Device model equivalent to Django Device model"""

    __tablename__ = "devices"

    # Basic device information
    name: Optional[str] = Field(max_length=256, description="Friendly identifier")
    node_type: NodeType = Field(default=NodeType.PEER)
    management_ip: Optional[str] = Field(description="Management IP address")
    router_id: Optional[str] = Field(description="BGP/OSPF Router ID")
    asn: Optional[int] = Field(description="BGP AS Number", ge=1, le=4294967295)

    # Hardware/System information
    serial_number: Optional[str] = Field(max_length=200, unique=True)
    stream: Optional[str] = Field(max_length=100)
    hostname: str = Field(max_length=256, unique=True)
    fqdn: Optional[str] = Field(max_length=256, unique=True)
    manufacturer: Optional[str] = Field(max_length=256)
    model: Optional[str] = Field(max_length=256)
    firmware: Optional[str] = Field(max_length=256)
    domain: Optional[str] = Field(max_length=256)
    cpu_arch: Optional[str] = Field(max_length=256)
    os_codename: Optional[str] = Field(max_length=256)
    product_name: Optional[str] = Field(max_length=256)
    processor_type: Optional[str] = Field(max_length=256)
    num_cpus: Optional[int]
    core_count: Optional[int]
    ram_size: Optional[int]

    # Status and operational fields
    is_active: bool = Field(default=True)
    last_checked: Optional[str]  # ISO format datetime string
    status: DeviceStatus = Field(default=DeviceStatus.OPERATIONAL)
    status_text: str = Field(
        default="All systems operational",
        max_length=500,
        description="Human readable description of the device status",
    )

    # Relationships
    interfaces: List["Interface"] = Relationship(back_populates="device")
    local_tunnels: List["Tunnel"] = Relationship(
        back_populates="local_node",
        sa_relationship_kwargs={"foreign_keys": "Tunnel.local_node_id"},
    )
    remote_tunnels: List["Tunnel"] = Relationship(
        back_populates="remote_node",
        sa_relationship_kwargs={"foreign_keys": "Tunnel.remote_node_id"},
    )


class Interface(BaseModel, table=True):
    """Interface model equivalent to Django Interface model"""

    __tablename__ = "interfaces"

    # Device relationship
    device_id: Optional[int] = Field(foreign_key="devices.id")
    device: Optional[Device] = Relationship(back_populates="interfaces")

    # Interface classification
    is_management: bool = Field(
        default=False, description="Is this interface used for management purposes?"
    )
    is_wan: bool = Field(
        default=False, description="Is this interface used for WAN connectivity?"
    )
    is_lan: bool = Field(
        default=False, description="Is this interface used for LAN connectivity?"
    )
    is_primary: bool = Field(
        default=False,
        description="Is this the primary interface of its type (WAN or LAN)?",
    )

    # Interface details
    name: str = Field(max_length=256)
    type: InterfaceType = Field(
        default=InterfaceType.PHYSICAL,
        description="Type of interface (e.g. physical, vlan, loopback, tunnel, bridge)",
    )
    up: bool = Field(default=False)
    hwaddr: Optional[str] = Field(max_length=50, description="Hardware address")
    ip_address: Optional[str] = Field(description="IP address")
    subnet_mask: Optional[int] = Field(ge=1, le=32, description="Subnet mask CIDR")
    vlan_id: Optional[int] = Field(ge=1, le=4094, description="VLAN ID")
    mtu: int = Field(default=1500)

    # VPP specific fields
    vpp_used: bool = Field(default=False, description="Is this interface used by VPP?")
    vpp_interface_name: Optional[str] = Field(
        max_length=256, description="VPP interface name"
    )
    is_enabled: bool = Field(default=True)
    description: Optional[str]

    # Relationships
    local_tunnels: List["Tunnel"] = Relationship(
        back_populates="local_interface",
        sa_relationship_kwargs={"foreign_keys": "Tunnel.local_interface_id"},
    )
    remote_tunnels: List["Tunnel"] = Relationship(
        back_populates="remote_interface",
        sa_relationship_kwargs={"foreign_keys": "Tunnel.remote_interface_id"},
    )
    local_wan_tunnels: List["Tunnel"] = Relationship(
        back_populates="local_wan_interface",
        sa_relationship_kwargs={"foreign_keys": "Tunnel.local_wan_interface_id"},
    )
    remote_wan_tunnels: List["Tunnel"] = Relationship(
        back_populates="remote_wan_interface",
        sa_relationship_kwargs={"foreign_keys": "Tunnel.remote_wan_interface_id"},
    )


# Forward reference resolution
from .tunnels import Tunnel

Device.model_rebuild()
Interface.model_rebuild()



"""
API endpoints for telemetry and metrics data.
"""

from typing import Dict, Any, Optional, List
from fastapi import APIRouter, HTTPException, Query
from loguru import logger
import time
from datetime import datetime, timedelta

from vrouter_agent.enhanced_stream_processor import get_stream_processor
from vrouter_agent.utils import get_device_serial_number

router = APIRouter(prefix="/telemetry", tags=["telemetry"])


@router.get("/metrics/")
async def get_metrics():
    """
    Get current system metrics from the stream processor.
    
    Returns:
        Current metrics including processing statistics, performance data, and system health
    """
    try:
        processor = await get_stream_processor()
        metrics = processor.get_metrics()
        
        # Enhance metrics with additional computed values
        enhanced_metrics = {
            **metrics,
            "device_serial": get_device_serial_number(),
            "timestamp": time.time(),
            "uptime_seconds": time.time() - metrics.get('start_time', time.time()),
        }
        
        # Calculate success rate
        total_processed = metrics.get('total_processed', 0)
        successful_processed = metrics.get('successful_processed', 0)
        if total_processed > 0:
            enhanced_metrics['success_rate'] = successful_processed / total_processed
        else:
            enhanced_metrics['success_rate'] = 0.0
        
        # Calculate failure rate
        failed_processed = metrics.get('failed_processed', 0)
        if total_processed > 0:
            enhanced_metrics['failure_rate'] = failed_processed / total_processed
        else:
            enhanced_metrics['failure_rate'] = 0.0
        
        logger.info("Retrieved system metrics")
        return enhanced_metrics
        
    except Exception as e:
        logger.error(f"Error retrieving metrics: {e}")
        raise HTTPException(status_code=500, detail="Failed to retrieve metrics")


@router.get("/status/")
async def get_system_status():
    """
    Get current system status and health information.
    
    Returns:
        System status including processor state, worker information, and health indicators
    """
    try:
        processor = await get_stream_processor()
        status = processor.get_status()
        
        # Enhance status with additional information
        enhanced_status = {
            **status,
            "device_serial": get_device_serial_number(),
            "timestamp": time.time(),
            "system_health": "healthy" if status.get("running", False) else "unhealthy",
        }
        
        # Add worker utilization status
        worker_utilization = status.get('worker_utilization', 0.0)
        if worker_utilization > 0.9:
            enhanced_status['worker_status'] = "high_load"
        elif worker_utilization > 0.7:
            enhanced_status['worker_status'] = "moderate_load"
        else:
            enhanced_status['worker_status'] = "normal"
        
        # Add queue status
        queue_size = status.get('queue_size', 0)
        if queue_size > 100:
            enhanced_status['queue_status'] = "high"
        elif queue_size > 50:
            enhanced_status['queue_status'] = "moderate"
        else:
            enhanced_status['queue_status'] = "normal"
        
        logger.info("Retrieved system status")
        return enhanced_status
        
    except Exception as e:
        logger.error(f"Error retrieving system status: {e}")
        raise HTTPException(status_code=500, detail="Failed to retrieve system status")


@router.get("/health/")
async def get_health_check():
    """
    Comprehensive health check endpoint.
    
    Returns:
        Detailed health information including all subsystems
    """
    try:
        processor = await get_stream_processor()
        metrics = processor.get_metrics()
        status = processor.get_status()
        
        current_time = time.time()
        
        # Determine overall health
        is_healthy = True
        health_issues = []
        
        # Check if processor is running
        if not status.get("running", False):
            is_healthy = False
            health_issues.append("Stream processor is not running")
        
        # Check worker utilization
        worker_utilization = status.get('worker_utilization', 0.0)
        if worker_utilization > 0.95:
            health_issues.append("High worker utilization")
        
        # Check queue size
        queue_size = status.get('queue_size', 0)
        if queue_size > 200:
            is_healthy = False
            health_issues.append("Queue size is critically high")
        elif queue_size > 100:
            health_issues.append("Queue size is elevated")
        
        # Check for recent activity
        last_activity = status.get('last_activity')
        if last_activity and (current_time - last_activity) > 300:  # 5 minutes
            health_issues.append("No recent processing activity")
        
        # Check failure rate
        total_processed = metrics.get('total_processed', 0)
        failed_processed = metrics.get('failed_processed', 0)
        if total_processed > 0:
            failure_rate = failed_processed / total_processed
            if failure_rate > 0.1:  # 10% failure rate
                is_healthy = False
                health_issues.append(f"High failure rate: {failure_rate:.2%}")
        
        health_status = {
            "healthy": is_healthy,
            "status": "healthy" if is_healthy else "unhealthy",
            "issues": health_issues,
            "checks": {
                "processor_running": status.get("running", False),
                "worker_utilization_ok": worker_utilization < 0.95,
                "queue_size_ok": queue_size < 200,
                "recent_activity": last_activity is not None and (current_time - last_activity) < 300,
                "failure_rate_ok": (failed_processed / max(total_processed, 1)) < 0.1,
            },
            "metrics_summary": {
                "total_processed": total_processed,
                "success_rate": (metrics.get('successful_processed', 0) / max(total_processed, 1)),
                "failure_rate": (failed_processed / max(total_processed, 1)),
                "average_processing_time": metrics.get('average_processing_time', 0),
                "queue_size": queue_size,
                "worker_utilization": worker_utilization,
            },
            "device_serial": get_device_serial_number(),
            "timestamp": current_time,
        }
        
        logger.info(f"Health check completed - Status: {health_status['status']}")
        return health_status
        
    except Exception as e:
        logger.error(f"Error during health check: {e}")
        raise HTTPException(status_code=500, detail="Failed to perform health check")


@router.get("/performance/")
async def get_performance_metrics():
    """
    Get detailed performance metrics.
    
    Returns:
        Performance-focused metrics and statistics
    """
    try:
        processor = await get_stream_processor()
        metrics = processor.get_metrics()
        status = processor.get_status()
        
        current_time = time.time()
        uptime = current_time - metrics.get('start_time', current_time)
        
        performance_metrics = {
            "processing_performance": {
                "total_processed": metrics.get('total_processed', 0),
                "successful_processed": metrics.get('successful_processed', 0),
                "failed_processed": metrics.get('failed_processed', 0),
                "retry_count": metrics.get('retry_count', 0),
                "average_processing_time_ms": metrics.get('average_processing_time', 0),
            },
            "system_performance": {
                "uptime_seconds": uptime,
                "uptime_hours": uptime / 3600,
                "worker_count": status.get('worker_count', 0),
                "worker_utilization": status.get('worker_utilization', 0.0),
                "queue_size": status.get('queue_size', 0),
            },
            "throughput": {
                "transactions_per_second": metrics.get('total_processed', 0) / max(uptime, 1),
                "successful_per_second": metrics.get('successful_processed', 0) / max(uptime, 1),
                "failed_per_second": metrics.get('failed_processed', 0) / max(uptime, 1),
            },
            "device_serial": get_device_serial_number(),
            "timestamp": current_time,
        }
        
        logger.info("Retrieved performance metrics")
        return performance_metrics
        
    except Exception as e:
        logger.error(f"Error retrieving performance metrics: {e}")
        raise HTTPException(status_code=500, detail="Failed to retrieve performance metrics")


@router.get("/system-info/")
async def get_system_info():
    """
    Get general system information.
    
    Returns:
        System information including device details and configuration
    """
    try:
        processor = await get_stream_processor()
        metrics = processor.get_metrics()
        
        system_info = {
            "device_serial": get_device_serial_number(),
            "processor_config": {
                "max_workers": getattr(processor, 'max_workers', 'unknown'),
                "batch_size": getattr(processor, 'batch_size', 'unknown'),
            },
            "runtime_info": {
                "start_time": metrics.get('start_time'),
                "start_time_iso": datetime.fromtimestamp(metrics.get('start_time', 0)).isoformat() if metrics.get('start_time') else None,
                "uptime_seconds": time.time() - metrics.get('start_time', time.time()),
                "last_activity": metrics.get('last_activity'),
                "last_activity_iso": datetime.fromtimestamp(metrics.get('last_activity', 0)).isoformat() if metrics.get('last_activity') else None,
            },
            "current_timestamp": time.time(),
            "current_timestamp_iso": datetime.now().isoformat(),
        }
        
        logger.info("Retrieved system information")
        return system_info
        
    except Exception as e:
        logger.error(f"Error retrieving system information: {e}")
        raise HTTPException(status_code=500, detail="Failed to retrieve system information")


@router.get("/alerts/")
async def get_alerts(
    severity: Optional[str] = Query(None, description="Filter by severity: low, medium, high, critical"),
):
    """
    Get current system alerts and warnings.
    
    Args:
        severity: Optional severity filter
    
    Returns:
        List of current alerts and warnings
    """
    try:
        processor = await get_stream_processor()
        metrics = processor.get_metrics()
        status = processor.get_status()
        
        alerts = []
        current_time = time.time()
        
        # Check for various alert conditions
        
        # Critical alerts
        if not status.get("running", False):
            alerts.append({
                "severity": "critical",
                "message": "Stream processor is not running",
                "timestamp": current_time,
                "component": "stream_processor"
            })
        
        queue_size = status.get('queue_size', 0)
        if queue_size > 200:
            alerts.append({
                "severity": "critical",
                "message": f"Queue size critically high: {queue_size}",
                "timestamp": current_time,
                "component": "queue"
            })
        
        # High severity alerts
        worker_utilization = status.get('worker_utilization', 0.0)
        if worker_utilization > 0.95:
            alerts.append({
                "severity": "high",
                "message": f"Worker utilization very high: {worker_utilization:.1%}",
                "timestamp": current_time,
                "component": "workers"
            })
        
        total_processed = metrics.get('total_processed', 0)
        failed_processed = metrics.get('failed_processed', 0)
        if total_processed > 0:
            failure_rate = failed_processed / total_processed
            if failure_rate > 0.1:
                alerts.append({
                    "severity": "high",
                    "message": f"High failure rate: {failure_rate:.1%}",
                    "timestamp": current_time,
                    "component": "processing"
                })
        
        # Medium severity alerts
        if queue_size > 100:
            alerts.append({
                "severity": "medium",
                "message": f"Queue size elevated: {queue_size}",
                "timestamp": current_time,
                "component": "queue"
            })
        
        if worker_utilization > 0.8:
            alerts.append({
                "severity": "medium",
                "message": f"Worker utilization high: {worker_utilization:.1%}",
                "timestamp": current_time,
                "component": "workers"
            })
        
        # Low severity alerts
        last_activity = status.get('last_activity')
        if last_activity and (current_time - last_activity) > 300:  # 5 minutes
            alerts.append({
                "severity": "low",
                "message": "No recent processing activity",
                "timestamp": current_time,
                "component": "activity"
            })
        
        # Filter by severity if requested
        if severity:
            alerts = [alert for alert in alerts if alert['severity'] == severity.lower()]
        
        # Sort by severity (critical first)
        severity_order = {"critical": 0, "high": 1, "medium": 2, "low": 3}
        alerts.sort(key=lambda x: severity_order.get(x['severity'], 4))
        
        result = {
            "alerts": alerts,
            "alert_count": len(alerts),
            "critical_count": len([a for a in alerts if a['severity'] == 'critical']),
            "high_count": len([a for a in alerts if a['severity'] == 'high']),
            "medium_count": len([a for a in alerts if a['severity'] == 'medium']),
            "low_count": len([a for a in alerts if a['severity'] == 'low']),
            "device_serial": get_device_serial_number(),
            "timestamp": current_time,
        }
        
        logger.info(f"Retrieved {len(alerts)} alerts")
        return result
        
    except Exception as e:
        logger.error(f"Error retrieving alerts: {e}")
        raise HTTPException(status_code=500, detail="Failed to retrieve alerts")



"""
Configuration Manager for Enhanced Stream Processing

This module provides configuration management for the enhanced stream processing
system with environment-specific settings and validation.
"""

import os
import yaml
from typing import Dict, Any, Optional
from pathlib import Path
from loguru import logger as log


class ConfigurationManager:
    """Manages configuration for enhanced stream processing."""
    
    def __init__(self, config_path: Optional[str] = None):
        """
        Initialize configuration manager.
        
        Args:
            config_path: Path to configuration file. If None, uses default path.
        """
        if config_path is None:
            config_path = self._get_default_config_path()
        
        self.config_path = config_path
        self.config = self._load_config()
        self.environment = self._detect_environment()
        
        # Apply environment-specific overrides
        self._apply_environment_config()
        
        log.info(f"Configuration loaded for environment: {self.environment}")
    
    def _get_default_config_path(self) -> str:
        """Get the default configuration file path."""
        base_dir = Path(__file__).parent.parent
        return str(base_dir / "config" / "enhanced_stream_config.yml")
    
    def _load_config(self) -> Dict[str, Any]:
        """Load configuration from YAML file."""
        try:
            with open(self.config_path, 'r') as file:
                config = yaml.safe_load(file)
                log.info(f"Configuration loaded from {self.config_path}")
                return config
        except FileNotFoundError:
            log.warning(f"Configuration file not found: {self.config_path}")
            return self._get_default_config()
        except yaml.YAMLError as e:
            log.error(f"Error parsing configuration file: {e}")
            return self._get_default_config()
    
    def _detect_environment(self) -> str:
        """Detect the current environment."""
        env = os.getenv('VROUTER_ENV', 'development').lower()
        valid_environments = ['development', 'staging', 'production']
        
        if env not in valid_environments:
            log.warning(f"Unknown environment '{env}', defaulting to 'development'")
            env = 'development'
        
        return env
    
    def _apply_environment_config(self) -> None:
        """Apply environment-specific configuration overrides."""
        if 'environments' not in self.config:
            return
        
        env_config = self.config['environments'].get(self.environment, {})
        if not env_config:
            return
        
        # Recursively merge environment config
        self._deep_merge(self.config, env_config)
        log.info(f"Applied {self.environment} environment configuration")
    
    def _deep_merge(self, base: Dict[str, Any], override: Dict[str, Any]) -> None:
        """Recursively merge configuration dictionaries."""
        for key, value in override.items():
            if key in base and isinstance(base[key], dict) and isinstance(value, dict):
                self._deep_merge(base[key], value)
            else:
                base[key] = value
    
    def _get_default_config(self) -> Dict[str, Any]:
        """Get default configuration when file is not available."""
        return {
            'worker_pool': {
                'max_workers': 5,
                'batch_size': 10,
                'timeout_seconds': 300
            },
            'retry_policy': {
                'max_retries': 3,
                'base_delay': 2.0,
                'max_delay': 30.0,
                'exponential_backoff': True
            },
            'monitoring': {
                'enable_metrics': True,
                'metrics_update_interval': 10.0,
                'log_level': 'INFO',
                'alerts': {
                    'high_queue_size': 100,
                    'high_error_rate': 0.1,
                    'low_throughput': 10,
                    'worker_failure_count': 3
                }
            },
            'performance': {
                'batch_processing_interval': 1.0,
                'enable_batch_processing': True,
                'queue_size_limit': 1000
            },
            'features': {
                'enable_enhanced_processing': True,
                'enable_legacy_fallback': True,
                'enable_metrics_export': True,
                'enable_health_checks': True
            }
        }
    
    def get(self, key_path: str, default: Any = None) -> Any:
        """
        Get configuration value using dot notation.
        
        Args:
            key_path: Configuration key path (e.g., 'worker_pool.max_workers')
            default: Default value if key not found
            
        Returns:
            Configuration value or default
        """
        keys = key_path.split('.')
        value = self.config
        
        try:
            for key in keys:
                value = value[key]
            return value
        except (KeyError, TypeError):
            return default
    
    def get_worker_config(self) -> Dict[str, Any]:
        """Get worker pool configuration."""
        return self.get('worker_pool', {})
    
    def get_retry_config(self) -> Dict[str, Any]:
        """Get retry policy configuration."""
        return self.get('retry_policy', {})
    
    def get_monitoring_config(self) -> Dict[str, Any]:
        """Get monitoring configuration."""
        return self.get('monitoring', {})
    
    def get_performance_config(self) -> Dict[str, Any]:
        """Get performance configuration."""
        return self.get('performance', {})
    
    def get_feature_flags(self) -> Dict[str, bool]:
        """Get feature flags."""
        return self.get('features', {})
    
    def is_feature_enabled(self, feature_name: str) -> bool:
        """Check if a feature is enabled."""
        return self.get_feature_flags().get(feature_name, False)
    
    def validate_config(self) -> bool:
        """
        Validate configuration values.
        
        Returns:
            True if configuration is valid, False otherwise
        """
        errors = []
        
        # Validate worker pool config
        max_workers = self.get('worker_pool.max_workers')
        if not isinstance(max_workers, int) or max_workers < 1:
            errors.append("worker_pool.max_workers must be a positive integer")
        
        batch_size = self.get('worker_pool.batch_size')
        if not isinstance(batch_size, int) or batch_size < 1:
            errors.append("worker_pool.batch_size must be a positive integer")
        
        # Validate retry config
        max_retries = self.get('retry_policy.max_retries')
        if not isinstance(max_retries, int) or max_retries < 0:
            errors.append("retry_policy.max_retries must be a non-negative integer")
        
        base_delay = self.get('retry_policy.base_delay')
        if not isinstance(base_delay, (int, float)) or base_delay < 0:
            errors.append("retry_policy.base_delay must be a non-negative number")
        
        # Validate monitoring config
        metrics_interval = self.get('monitoring.metrics_update_interval')
        if not isinstance(metrics_interval, (int, float)) or metrics_interval <= 0:
            errors.append("monitoring.metrics_update_interval must be a positive number")
        
        if errors:
            for error in errors:
                log.error(f"Configuration validation error: {error}")
            return False
        
        log.info("Configuration validation passed")
        return True
    
    def reload_config(self) -> bool:
        """
        Reload configuration from file.
        
        Returns:
            True if reload successful, False otherwise
        """
        try:
            old_config = self.config.copy()
            self.config = self._load_config()
            self._apply_environment_config()
            
            if self.validate_config():
                log.info("Configuration reloaded successfully")
                return True
            else:
                # Restore old config if validation fails
                self.config = old_config
                log.error("Configuration reload failed validation, keeping old config")
                return False
                
        except Exception as e:
            log.error(f"Error reloading configuration: {e}")
            return False
    
    def get_environment_info(self) -> Dict[str, Any]:
        """Get information about the current environment."""
        return {
            'environment': self.environment,
            'config_path': self.config_path,
            'features_enabled': [
                feature for feature, enabled in self.get_feature_flags().items()
                if enabled
            ],
            'worker_count': self.get('worker_pool.max_workers'),
            'batch_size': self.get('worker_pool.batch_size'),
            'retry_max': self.get('retry_policy.max_retries')
        }


# Global configuration instance
_config_manager = None


def get_config_manager() -> ConfigurationManager:
    """Get the global configuration manager instance."""
    global _config_manager
    if _config_manager is None:
        _config_manager = ConfigurationManager()
    return _config_manager


def reload_config() -> bool:
    """Reload the global configuration."""
    global _config_manager
    if _config_manager is not None:
        return _config_manager.reload_config()
    return False



import os
import json
from typing import Optional, Dict, List, Any, Union
from contextlib import contextmanager
from dataclasses import dataclass
from vrouter_agent.utils import run_command
import mcrpc
import re
import time
from datetime import datetime
from cryptography.fernet import Fernet
from vrouter_agent.core.config import settings

from loguru import logger as log


def get_multichain_client(chain_name: str, user: str):
    """Multichain client using secure directory structure

    Args:
        chain_name (str): chain_name
        user (str): user (maintained for compatibility but uses secure location)

    Raises:
        OSError: File error

    Returns:
        client: multichain_client
    """
    try:
        # Use the secure multichain directory
        multichain_dir = f"/var/lib/multichain/.multichain/{chain_name}"
        if not os.path.exists(multichain_dir):
            raise OSError(
                f"Multichain {chain_name} is not connected to this node. Please try again."
            )
        rpcuser = None
        rpcpasswd = None

        with open(
            os.path.join(multichain_dir, "multichain.conf"),
            "r",
            encoding="utf-8",
        ) as blockchain_file:
            for line in blockchain_file.readlines():
                if line.startswith("rpcuser"):
                    rpcuser = line.split("=")[1].strip()
                if line.startswith("rpcpassword"):
                    rpcpasswd = line.split("=")[1].strip()
        with open(
            os.path.join(multichain_dir, "params.dat"),
            "r",
            encoding="utf-8",
        ) as blockchain_file:
            for line in blockchain_file.readlines():
                if line.startswith("default-rpc-port"):
                    rpcport = re.findall(r"\d+", line)[0]


        client = mcrpc.RpcClient("localhost", rpcport, rpcuser, rpcpasswd, False)
        return client

    except IOError as error:
        log.exception("Error occurred: " + str(error))

    except KeyError as error:
        log.exception(str(error))

    return None


@dataclass
class MultichainConfig:
    """Configuration for multichain connection."""
    chain_name: str
    user: str
    rpc_host: str = "localhost"
    max_retries: int = 3
    retry_delay: float = 5.0
    timeout: float = 30.0


@dataclass
class StreamItem:
    """Represents a stream item from multichain."""
    txid: str
    data: str

    publishers: Optional[List[str]] = None
    keys: Optional[List[str]] = None
    offchain: Optional[bool] = False
    available: Optional[bool] = True
    confirmations: Optional[int] = 0
    blocktime: Optional[int] = None
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'StreamItem':
        """Create StreamItem from dictionary."""
        return cls(
            txid=data.get('txid', ''),
            publishers=data.get('publishers', []),
            keys=data.get('keys', []),
            data=data.get('data', {}),
            confirmations=data.get('confirmations', 0),
            blocktime=data.get('blocktime'),
            offchain=data.get('offchain', False),
            available=data.get('available', True)
        )
    @classmethod
    def get_decrypted_data(self) -> str:
        """Decrypt the data using the private key"""
        # convert data to bytes
        data = bytes.fromhex(self.data)
        # decrypt the data
        key = Fernet(settings.config.global_.secret_key.encode("utf-8"))
        
        decrypted_data = key.decrypt(data)
        return decrypted_data.decode("utf-8")




class MultichainConnectionError(Exception):
    """Exception raised when multichain connection fails."""
    pass


class StreamNotFoundError(Exception):
    """Exception raised when stream is not found."""
    pass


@contextmanager
def multichain_client_context(config: MultichainConfig):
    """Context manager for multichain client connections."""
    client = None
    try:
        client = get_multichain_client_enhanced(config)
        if not client:
            raise MultichainConnectionError(f"Failed to connect to chain: {config.chain_name}")
        yield client
    finally:
        # Cleanup if needed
        if client:
            log.debug(f"Closed multichain client connection for {config.chain_name}")


def get_multichain_client_enhanced(config: MultichainConfig) -> Optional[mcrpc.RpcClient]:
    """Enhanced multichain client with better error handling and configuration.
    
    Args:
        config: Multichain configuration
        
    Returns:
        Optional[mcrpc.RpcClient]: Multichain client or None if failed
        
    Raises:
        MultichainConnectionError: If connection fails after retries
    """
    for attempt in range(config.max_retries):
        try:
            log.info(f"Attempting to connect to multichain {config.chain_name} (attempt {attempt + 1}/{config.max_retries})")
            
            # Use the secure multichain directory
            multichain_dir = f"/var/lib/multichain/.multichain/{config.chain_name}"
            if not os.path.exists(multichain_dir):
                raise OSError(
                    f"Multichain {config.chain_name} is not connected to this node. Please try again."
                )
            
            # Read RPC credentials
            rpc_config = _read_rpc_config(multichain_dir)
            
            # Test connection with getinfo
            _test_multichain_connection(config.chain_name, config.user)
            
            # Create client
            client = mcrpc.RpcClient(
                config.rpc_host, 
                rpc_config['rpcport'], 
                rpc_config['rpcuser'], 
                rpc_config['rpcpasswd'], 
                False
            )
            
            # Test client connection
            try:
                client.getinfo()
                log.info(f"Successfully connected to multichain {config.chain_name}")
                return client
            except Exception as e:
                log.warning(f"Client connection test failed: {e}")
                raise
                
        except Exception as error:
            log.warning(f"Connection attempt {attempt + 1} failed: {error}")
            if attempt < config.max_retries - 1:
                log.info(f"Retrying in {config.retry_delay} seconds...")
                time.sleep(config.retry_delay)
            else:
                log.error(f"All connection attempts failed for {config.chain_name}")
                raise MultichainConnectionError(f"Failed to connect after {config.max_retries} attempts")
    
    return None


def _read_rpc_config(multichain_dir: str) -> Dict[str, str]:
    """Read RPC configuration from multichain directory."""
    rpcuser = None
    rpcpasswd = None
    rpcport = None
    
    # Read multichain.conf
    conf_file = os.path.join(multichain_dir, "multichain.conf")
    if os.path.exists(conf_file):
        with open(conf_file, "r", encoding="utf-8") as f:
            for line in f.readlines():
                line = line.strip()
                if line.startswith("rpcuser"):
                    rpcuser = line.split("=", 1)[1].strip()
                elif line.startswith("rpcpassword"):
                    rpcpasswd = line.split("=", 1)[1].strip()
    
    # Read params.dat for port
    params_file = os.path.join(multichain_dir, "params.dat")
    if os.path.exists(params_file):
        with open(params_file, "r", encoding="utf-8") as f:
            for line in f.readlines():
                if line.startswith("default-rpc-port"):
                    rpcport = re.findall(r"\d+", line)[0]
    
    if not all([rpcuser, rpcpasswd, rpcport]):
        raise ValueError("Missing RPC configuration parameters")
    
    return {
        'rpcuser': rpcuser,
        'rpcpasswd': rpcpasswd,
        'rpcport': rpcport
    }


def _test_multichain_connection(chain_name: str, user: str) -> None:
    """Test multichain connection using direct RPC instead of CLI.
    
    This method uses direct RPC connection to test multichain availability,
    avoiding sudo requirements and working with the secure multichain setup.
    """
    try:
        # Use the secure multichain directory
        multichain_dir = f"/var/lib/multichain/.multichain/{chain_name}"
        
        if not os.path.exists(multichain_dir):
            raise MultichainConnectionError(
                f"Multichain {chain_name} directory not found at {multichain_dir}"
            )
        
        # Read RPC configuration from secure location
        rpc_config = _read_rpc_config(multichain_dir)
        
        # Test direct RPC connection instead of CLI
        client = mcrpc.RpcClient(
            "localhost",
            rpc_config['rpcport'],
            rpc_config['rpcuser'],
            rpc_config['rpcpasswd'],
            False
        )
        
        # Test the connection with a simple getinfo call
        info = client.getinfo()
        log.debug(f"MultiChain connection test successful for {chain_name}: {info.get('version', 'unknown version')}")
        
    except Exception as e:
        raise MultichainConnectionError(f"RPC connection test failed for {chain_name}: {str(e)}")


def list_stream_items_enhanced(
    config: MultichainConfig,
    stream_name: str,
    start: Optional[str] = None,
    count: int = 100,
    verbose: bool = True,
    skip_txids: Optional[List[str]] = None
) -> List[StreamItem]:
    """Enhanced function to list stream items with better filtering and pagination.
    
    Args:
        config: Multichain configuration
        stream_name: Name of the stream
        start: Starting point for pagination
        count: Maximum number of items to return
        verbose: Whether to include verbose information
        skip_txids: List of transaction IDs to skip
        
    Returns:
        List[StreamItem]: List of stream items
        
    Raises:
        StreamNotFoundError: If stream doesn't exist
        MultichainConnectionError: If connection fails
    """
    skip_txids = skip_txids or []
    
    with multichain_client_context(config) as client:
        # Ensure stream subscription
        ensure_stream_subscription(client, stream_name)
        
        # Get stream items
        try:
            raw_items = client.liststreamitems(stream_name, verbose, count, start)
        except Exception as e:
            if "stream not found" in str(e).lower():
                raise StreamNotFoundError(f"Stream '{stream_name}' not found")
            raise MultichainConnectionError(f"Failed to list stream items: {e}")
        
        # Convert to StreamItem objects and filter
        items = []
        for raw_item in raw_items:
            if raw_item.get('txid') not in skip_txids:
                try:
                    item = StreamItem.from_dict(raw_item)
                    items.append(item)
                except Exception as e:
                    log.warning(f"Failed to parse stream item: {e}")
                    continue
        
        log.info(f"Retrieved {len(items)} items from stream '{stream_name}'")
        return items


def ensure_stream_subscription(client: mcrpc.RpcClient, stream_name: str) -> bool:
    """Ensure a stream is subscribed.
    
    Args:
        client: Multichain RPC client
        stream_name: Name of the stream
        
    Returns:
        bool: True if subscribed successfully
        
    Raises:
        StreamNotFoundError: If stream doesn't exist
    """
    try:
        streams = client.liststreams(stream_name)
        
        if not streams:
            raise StreamNotFoundError(f"Stream '{stream_name}' not found")
        
        for stream in streams:
            if not stream.get("subscribed", False):
                log.info(f"Subscribing to stream: {stream['name']}")
                client.subscribe(stream["name"])
                return True
        
        log.debug(f"Already subscribed to stream: {stream_name}")
        return True
        
    except Exception as e:
        if "stream not found" in str(e).lower():
            raise StreamNotFoundError(f"Stream '{stream_name}' not found")
        raise


def publish_to_stream(
    config: MultichainConfig,
    stream_name: str,
    key: str,
    data: Union[str, Dict[str, Any]],
    options: Optional[Dict[str, Any]] = None
) -> str:
    """Publish data to a multichain stream.
    
    Args:
        config: Multichain configuration
        stream_name: Name of the stream
        key: Key for the data
        data: Data to publish (string or dictionary)
        options: Additional options for publishing
        
    Returns:
        str: Transaction ID of the published item
        
    Raises:
        StreamNotFoundError: If stream doesn't exist
        MultichainConnectionError: If connection fails
    """
    with multichain_client_context(config) as client:
        # Ensure stream exists and we're subscribed
        ensure_stream_subscription(client, stream_name)
        
        # Convert data to hex if it's a dictionary
        if isinstance(data, dict):
            data_str = json.dumps(data)
            data_hex = data_str.encode('utf-8').hex()
        else:
            data_hex = data.encode('utf-8').hex()
        
        try:
            # Publish to stream
            if options:
                txid = client.publishfrom(
                    client.getaddresses()[0],  # Use first address
                    stream_name,
                    key,
                    data_hex,
                    options
                )
            else:
                txid = client.publish(stream_name, key, data_hex)
            
            log.info(f"Published to stream '{stream_name}' with key '{key}', txid: {txid}")
            return txid
            
        except Exception as e:
            if "stream not found" in str(e).lower():
                raise StreamNotFoundError(f"Stream '{stream_name}' not found")
            raise MultichainConnectionError(f"Failed to publish to stream: {e}")


def get_stream_info(config: MultichainConfig, stream_name: str) -> Dict[str, Any]:
    """Get information about a stream.
    
    Args:
        config: Multichain configuration
        stream_name: Name of the stream
        
    Returns:
        Dict[str, Any]: Stream information
        
    Raises:
        StreamNotFoundError: If stream doesn't exist
    """
    with multichain_client_context(config) as client:
        try:
            streams = client.liststreams(stream_name, True)  # verbose=True
            
            if not streams:
                raise StreamNotFoundError(f"Stream '{stream_name}' not found")
            
            return streams[0]
            
        except Exception as e:
            if "stream not found" in str(e).lower():
                raise StreamNotFoundError(f"Stream '{stream_name}' not found")
            raise MultichainConnectionError(f"Failed to get stream info: {e}")


def list_all_streams(config: MultichainConfig) -> List[Dict[str, Any]]:
    """List all streams on the multichain.
    
    Args:
        config: Multichain configuration
        
    Returns:
        List[Dict[str, Any]]: List of stream information
    """
    with multichain_client_context(config) as client:
        try:
            return client.liststreams("*", True)  # All streams, verbose=True
        except Exception as e:
            raise MultichainConnectionError(f"Failed to list streams: {e}")


def create_stream(
    config: MultichainConfig,
    stream_name: str,
    open_flag: bool = True,
    options: Optional[Dict[str, Any]] = None
) -> str:
    """Create a new stream on the multichain.
    
    Args:
        config: Multichain configuration
        stream_name: Name of the new stream
        open_flag: Whether the stream is open for writing
        options: Additional stream options
        
    Returns:
        str: Transaction ID of the stream creation
    """
    with multichain_client_context(config) as client:
        try:
            if options:
                txid = client.create("stream", stream_name, open_flag, options)
            else:
                txid = client.create("stream", stream_name, open_flag)
            
            log.info(f"Created stream '{stream_name}' with txid: {txid}")
            return txid
            
        except Exception as e:
            raise MultichainConnectionError(f"Failed to create stream: {e}")


def list_stream_items(
    config: MultichainConfig,
    stream_name: str,
    start: Optional[str] = None,
    count: int = 100,
    verbose: bool = True
) -> List[StreamItem]:
    """Compatibility wrapper for list_stream_items_enhanced.
    
    This function provides backward compatibility for existing code that expects
    the list_stream_items function. It calls list_stream_items_enhanced with
    default parameters.
    
    Args:
        config: Multichain configuration
        stream_name: Name of the stream
        start: Starting point for pagination
        count: Maximum number of items to return
        verbose: Whether to include verbose information
        
    Returns:
        List[StreamItem]: List of stream items
        
    Raises:
        StreamNotFoundError: If stream doesn't exist
        MultichainConnectionError: If connection fails
    """
    return list_stream_items_enhanced(
        config=config,
        stream_name=stream_name,
        start=start,
        count=count,
        verbose=verbose
    )


def get_multichain_status(config: MultichainConfig) -> Dict[str, Any]:
    """Get comprehensive status of the multichain node.
    
    Args:
        config: Multichain configuration
        
    Returns:
        Dict[str, Any]: Status information
    """
    with multichain_client_context(config) as client:
        try:
            info = client.getinfo()
            
            # Get additional information
            peer_info = client.getpeerinfo()
            blockchain_info = client.getblockchaininfo()
            
            return {
                "node_info": info,
                "peer_count": len(peer_info),
                "peers": peer_info,
                "blockchain": blockchain_info,
                "timestamp": datetime.now().isoformat(),
                "chain_name": config.chain_name
            }
            
        except Exception as e:
            raise MultichainConnectionError(f"Failed to get multichain status: {e}")




from fastapi import Depends, HTTPException
from sqlmodel import Session, select
from vrouter_agent.core.db import get_session
from vrouter_agent.utils.config import get_device_serial_number
from vrouter_agent.models import Order

from fastapi import APIRouter

router = APIRouter()

@router.get("/")
async def get_orders(
    session: Session = Depends(get_session),
    offset: int = 0,
    limit: int = 10,
):
    items = session.exec(select(Order).offset(offset).limit(limit)).all()
    # Check if any orders exist
    if not items:
        return HTTPException(status_code=404, detail="No orders found")
    host = get_device_serial_number()

    return [
        {
            "id": item.id,
            "order_number": item.order_number,
            "type": item.type,
            "status": item.status,
            "tunnels": item.tunnels_in_scope(
                host
            ),  # only return the tunnels in scope for the host
        }
        for item in items
    ]


@router.get("/{order_id}")
async def get_order_by_id(
    order_id: int,
    session: Session = Depends(get_session),
):
    item = session.get(Order, order_id)
    if not item:
        return HTTPException(status_code=404, detail="Order not found")
    host = get_device_serial_number()

    return {
        "id": item.id,
        "order_number": item.order_number,
        "type": item.type,
        "status": item.status,
        "tunnels": item.tunnels_in_scope(
            host
        ),  # only return the tunnels in scope for the host
    }







# from vrouter_agent.utils.helpers import run_frr_command
# from typing import List, Optional
# from loguru import logger as log
# from vrouter_agent.core.models import OSPFConfig
# from pydantic import BaseModel
# from typing import List, Optional
# from vrouter_agent.core.config import settings 
# from vrouter_agent.core.base import CreatedWireguardTunnel



# class FRR:
#     container_name: str = "frr"
#     namespace: str = "frr"
#     namespace_dir: str = "/run/netns/frr"
#     container_image: str = "public.ecr.aws/v3c4v9y4/frr:latest"

#     def __init__(
#         self,
#         ospf_config: OSPFConfig,
#         wireguard_tunnels: List[CreatedWireguardTunnel],
#     ):
#         self.ospf_config = ospf_config
#         self.wireguard_tunnels = wireguard_tunnels
#         self.route_reflector = settings.config.route_reflector
#         self.route_reflector_controller = settings.config.route_reflector_controller
#         self.vpp_lan_interface = settings.config.interfaces["lan_primary"]

#     def _run_frr_command(self, commands: List[str]):
#         """Helper method to run FRR commands."""
#         command_string = "\n".join(commands)
#         run_frr_command(command_string, self.container_name)

#     def _configure_interface(
#         self, interface_name: str, ip_address: str, ospf_area: Optional[str] = None
#     ):
#         """Helper method to configure an interface."""
#         commands = [
#             "enable",
#             "configure terminal",
#             f"interface {interface_name}",
#             f"ip address {ip_address}",
#         ]

#         if ospf_area:
#             commands += [
#                 f"ip ospf 1 area {ospf_area}",
#                 "ip ospf network point-to-point",
#                 "ip ospf hello-interval 2",
#                 "ip ospf dead-interval 8",
#                 "ip ospf cost 10",
#             ]

#         if self._run_frr_command(commands):
#             log.info(f"Added interface {interface_name} with IP address {ip_address}")

#     def add_loopback_interface(self):
#         """Add a loopback interface."""
#         self._configure_interface("lo", self.route_reflector.loopback + "/24")
#         # add ip ospf 2 area 0.0.0.0
#         commands = [
#             "enable",
#             "configure terminal",
#             "interface lo",
#             f"ip ospf 2 area 0.0.0.0",
#         ]
#         self._run_frr_command(commands)

#         log.info(
#             f"Added loopback interface with IP address {self.route_reflector.loopback}"
#         )

#     def add_ospf_config(self):
#         """Add core OSPF configuration."""
#         commands = [
#             "enable",
#             "configure terminal",
#             "router ospf 2",
#             f"router-id {self.ospf_config.router_id}",
#             "exit",
#         ]
#         self._run_frr_command(commands)

#     def add_address_family(self):
#         """Add BGP address-family configuration."""

#         def get_lan_network(ip_addr: str) -> str:
#             ip_parts = ip_addr.split(".")
#             ip_parts[-1] = "0"
#             return ".".join(ip_parts) + "/24"

#         commands = [
#             "enable",
#             "configure terminal",
#             "router bgp 64512",
#             "address-family ipv4 unicast",
#             f"network {get_lan_network(self.vpp_lan_interface.ip_address)}",
#             "redistribute ospf 1",
#             "exit-address-family",
#         ]
#         self._run_frr_command(commands)

#     def add_core_bgp_config(self):
#         """Add core BGP configuration including BGP neighbor."""
#         commands = [
#             "enable",
#             "configure terminal",
#             "router bgp 64512",
#             f"bgp router-id {self.route_reflector.loopback}",
#             f"neighbor {self.route_reflector_controller.loopback.split('/')[0]} remote-as internal",
#             f"neighbor {self.route_reflector_controller.loopback.split('/')[0]} update-source lo",
#             "neighbor "
#             + self.route_reflector_controller.loopback.split("/")[0]
#             + " timers connect 10",
#         ]
#         self._run_frr_command(commands)

#     def add_client_ospf_config(self):
#         """Add client OSPF configuration."""

#         self._configure_interface(
#             self.vpp_lan_interface.interface_name,
#             f"{self.vpp_lan_interface.ip_address}/{self.vpp_lan_interface.prefix_len}",
#             self.ospf_config.client_router_area,
#         )

#         commands = [
#             "enable",
#             "configure terminal",
#             "router ospf 1",
#             f"router-id {self.ospf_config.client_router_id}",
#             "redistribute bgp",
#             "exit",
#         ]
#         self._run_frr_command(commands)

#     def add_wireguard_config(
#         self,
#         interface_name: str,
#         ip_addr: str,
#     ):
#         """Add WireGuard interface configuration."""
#         self._configure_interface(interface_name, ip_addr)

#         commands = [
#             "enable",
#             "configure terminal",
#             f"interface {interface_name}",
#             "ip ospf 2 area 0.0.0.0",
#             "ip ospf hello-interval 2",
#             "ip ospf dead-interval 8",
#             "ip ospf cost 10",
#             "ip ospf network point-to-point",
#             "exit",
#         ]
#         self._run_frr_command(commands)

#     def apply_all_configs(self):
#         self.add_loopback_interface()
#         self.add_ospf_config()
#         self.add_address_family()
#         self.add_core_bgp_config()
#         if self.ospf_config.client_enabled:
#             self.add_client_ospf_config()
#         for wg in self.wireguard_tunnels:
#             self.add_wireguard_config(wg.name, wg.ip_address)
#         log.info("All FRR configurations applied successfully!")

#     def delete_all_configs(self):
#         """Delete all FRR configurations."""

#         # delete core OSPF configuration
#         commands = [
#             "enable",
#             "configure terminal",
#             "no router ospf 2",
#             "no router ospf 1",
#             "exit",
#         ]
#         self._run_frr_command(commands)

#         log.info("All FRR configurations deleted successfully!")



from typing import List, Optional, Dict, Any, Union
from loguru import logger as log
from ipaddress import IPv4Address, IPv4Network, IPv4Interface
from pydantic import BaseModel

from vrouter_agent.utils.helpers import run_frr_command
from vrouter_agent.core.frr_config import (
    OspfConfig, OspfArea, OspfInterface, OspfRedistribution,
    BgpConfig, BgpPeer, BgpNetwork, BgpRedistribution,
    InterfaceBase, RouterId, FrrDeviceConfig
)
from vrouter_agent.core.config import settings
from vrouter_agent.core.base import CreatedWireguardTunnel


class FRRCommandResult(BaseModel):
    """Model for FRR command execution results"""
    success: bool
    output: str = ""
    error: str = ""


class FRR:
    """Enhanced FRR management class to handle comprehensive FRR configurations"""
    
    container_name: str = "frr"
    namespace: str = "frr"
    namespace_dir: str = "/run/netns/frr"
    container_image: str = "public.ecr.aws/v3c4v9y4/frr:latest"

    def __init__(
        self,
        device_config: Optional[FrrDeviceConfig] = None,
        ospf_config: Optional[OspfConfig] = None,
        bgp_config: Optional[BgpConfig] = None,
        interfaces: Optional[List[InterfaceBase]] = None,
        wireguard_tunnels: Optional[List[CreatedWireguardTunnel]] = None,
    ):
        """
        Initialize FRR with either a comprehensive device configuration or individual components.
        
        Args:
            device_config: Complete FRR device configuration
            ospf_config: OSPF configuration if not using device_config
            bgp_config: BGP configuration if not using device_config
            interfaces: List of interface configurations if not using device_config
            wireguard_tunnels: List of wireguard tunnel configurations
        """
        if device_config:
            self.device_config = device_config
            self.ospf_config = device_config.ospf
            self.bgp_config = device_config.bgp
            self.interfaces = device_config.interfaces
        else:
            self.ospf_config = ospf_config
            self.bgp_config = bgp_config
            self.interfaces = interfaces or []
            self.device_config = None
            
        self.wireguard_tunnels = wireguard_tunnels or []
        
        # Load settings for backward compatibility
        if hasattr(settings, 'config'):
            self.route_reflector = getattr(settings.config, 'route_reflector', None)
            self.route_reflector_controller = getattr(settings.config, 'route_reflector_controller', None)
            self.vpp_lan_interface = getattr(settings.config, 'interfaces', {}).get('lan_primary')
        else:
            self.route_reflector = None
            self.route_reflector_controller = None
            self.vpp_lan_interface = None

    def run_frr_command(self, commands: List[str]) -> FRRCommandResult:
        """
        Execute FRR commands with improved error handling and feedback.
        
        Args:
            commands: List of FRR commands to execute
            
        Returns:
            FRRCommandResult: Result object with success status and output/error
        """
        command_string = "\n".join(commands)
        
        try:
            output = run_frr_command(command_string, self.container_name)
            log.debug(f"FRR command executed successfully: {commands}")
            return FRRCommandResult(success=True, output=output)
        except Exception as e:
            error_msg = f"Failed to execute FRR command: {e}"
            log.error(error_msg)
            return FRRCommandResult(success=False, error=error_msg)

    def configure_interface(
        self, 
        interface_name: str, 
        ip_address: str, 
        ospf_area: Optional[str] = None,
        ospf_process_id: int = 1,
        hello_interval: int = 10,
        dead_interval: int = 40,
        cost: int = 10,
        network_type: str = "point-to-point",
        passive: bool = False,
        description: Optional[str] = None,
    ) -> FRRCommandResult:
        """
        Configure a network interface with enhanced options.
        
        Args:
            interface_name: Name of the interface to configure
            ip_address: IP address with CIDR notation (e.g., '192.168.1.1/24')
            ospf_area: OSPF area ID (if applicable)
            ospf_process_id: OSPF process ID
            hello_interval: OSPF hello interval in seconds
            dead_interval: OSPF dead interval in seconds
            cost: OSPF interface cost
            network_type: OSPF network type
            passive: Whether the interface should be passive in OSPF
            description: Interface description
            
        Returns:
            FRRCommandResult: Result of the configuration operation
        """
        commands = [
            "enable",
            "configure terminal",
            f"interface {interface_name}",
            f"ip address {ip_address}",
        ]
        
        if description:
            commands.append(f"description {description}")

        if ospf_area:
            commands.extend([
                f"ip ospf {ospf_process_id} area {ospf_area}",
                f"ip ospf network {network_type}",
                f"ip ospf hello-interval {hello_interval}",
                f"ip ospf dead-interval {dead_interval}",
                f"ip ospf cost {cost}",
            ])
            
            if passive:
                commands.append("ip ospf passive")

        result = self.run_frr_command(commands)
        if result.success:
            log.info(f"Configured interface {interface_name} with IP address {ip_address}")
        
        return result

    def add_loopback_interface(self, ip_address: str, ospf_area: str = "0.0.0.0", ospf_process_id: int = 1) -> FRRCommandResult:
        """
        Add and configure a loopback interface.
        
        Args:
            ip_address: IP address for the loopback interface
            ospf_area: OSPF area ID for the loopback
            ospf_process_id: OSPF process ID
            
        Returns:
            FRRCommandResult: Result of the operation
        """
        # First create the basic loopback interface
        result = self.configure_interface("lo", ip_address)
        if not result.success:
            return result
        
        # Then add OSPF configuration if needed
        if ospf_area:
            commands = [
                "enable",
                "configure terminal",
                "interface lo",
                f"ip ospf {ospf_process_id} area {ospf_area}",
            ]
            result = self.run_frr_command(commands)
            
        if result.success:
            log.info(f"Added loopback interface with IP address {ip_address}")
            
        return result

    def configure_ospf(self, config: OspfConfig) -> FRRCommandResult:
        """
        Configure OSPF with comprehensive options.
        
        Args:
            config: Complete OSPF configuration
            
        Returns:
            FRRCommandResult: Result of the operation
        """
        commands = [
            "enable",
            "configure terminal",
            f"router ospf {config.process_id}",
            f"router-id {config.router_id}",
        ]
        
        # Add reference bandwidth if specified
        if config.reference_bandwidth:
            commands.append(f"auto-cost reference-bandwidth {config.reference_bandwidth}")
            
        # Set passive interface default if specified
        if config.passive_interfaces_default:
            commands.append("passive-interface default")
            
        # Configure passive interfaces
        if config.passive_interfaces:
            for interface in config.passive_interfaces:
                commands.append(f"passive-interface {interface}")
                
        # Configure non-passive interfaces
        if config.no_passive_interfaces:
            for interface in config.no_passive_interfaces:
                commands.append(f"no passive-interface {interface}")
                
        # Configure redistribution
        if config.redistribute:
            for redist in config.redistribute:
                redist_cmd = f"redistribute {redist.protocol}"
                if redist.metric:
                    redist_cmd += f" metric {redist.metric}"
                if redist.metric_type:
                    redist_cmd += f" metric-type {redist.metric_type}"
                if redist.route_map:
                    redist_cmd += f" route-map {redist.route_map}"
                commands.append(redist_cmd)
                
        # Configure default information originate
        if config.default_information_originate:
            commands.append("default-information originate")
            
        commands.append("exit")
        
        result = self.run_frr_command(commands)
        if result.success:
            log.info(f"Configured OSPF process {config.process_id} with router ID {config.router_id}")
            
        return result

    def configure_bgp(self, config: BgpConfig) -> FRRCommandResult:
        """
        Configure BGP with comprehensive options.
        
        Args:
            config: Complete BGP configuration
            
        Returns:
            FRRCommandResult: Result of the operation
        """
        commands = [
            "enable",
            "configure terminal",
            f"router bgp {config.local_as}",
            f"bgp router-id {config.router_id}",
        ]
        
        # Configure BGP settings
        if config.log_neighbor_changes:
            commands.append("bgp log-neighbor-changes")
            
        if config.deterministic_med:
            commands.append("bgp deterministic-med")
            
        if config.always_compare_med:
            commands.append("bgp always-compare-med")
            
        # Configure BGP peers
        for peer in config.peers:
            peer_cmd = f"neighbor {peer.peer_ip} remote-as {peer.remote_as}"
            commands.append(peer_cmd)
            
            if peer.description:
                commands.append(f"neighbor {peer.peer_ip} description {peer.description}")
                
            if peer.update_source:
                commands.append(f"neighbor {peer.peer_ip} update-source {peer.update_source}")
                
            if peer.next_hop_self:
                commands.append(f"neighbor {peer.peer_ip} next-hop-self")
                
            if peer.ebgp_multihop:
                commands.append(f"neighbor {peer.peer_ip} ebgp-multihop {peer.ebgp_multihop}")
                
            if peer.password:
                commands.append(f"neighbor {peer.peer_ip} password {peer.password}")
                
            if peer.timers:
                keepalive = peer.timers.get("keepalive", 60)
                holdtime = peer.timers.get("holdtime", 180)
                commands.append(f"neighbor {peer.peer_ip} timers {keepalive} {holdtime}")
            
            # Address family specific configurations
            commands.append("address-family ipv4 unicast")
            
            if peer.route_map_in:
                commands.append(f"neighbor {peer.peer_ip} route-map {peer.route_map_in} in")
                
            if peer.route_map_out:
                commands.append(f"neighbor {peer.peer_ip} route-map {peer.route_map_out} out")
                
            if peer.prefix_list_in:
                commands.append(f"neighbor {peer.peer_ip} prefix-list {peer.prefix_list_in} in")
                
            if peer.prefix_list_out:
                commands.append(f"neighbor {peer.peer_ip} prefix-list {peer.prefix_list_out} out")
                
            commands.append("exit-address-family")
            
        # Configure networks
        if config.networks:
            commands.append("address-family ipv4 unicast")
            for network in config.networks:
                net_cmd = f"network {network.network}"
                if network.route_map:
                    net_cmd += f" route-map {network.route_map}"
                commands.append(net_cmd)
            
            # Configure redistribution
            if config.redistribute:
                for redist in config.redistribute:
                    redist_cmd = f"redistribute {redist.protocol}"
                    if redist.metric:
                        redist_cmd += f" metric {redist.metric}"
                    if redist.route_map:
                        redist_cmd += f" route-map {redist.route_map}"
                    commands.append(redist_cmd)
                    
            commands.append("exit-address-family")
            
        commands.append("exit")
        
        result = self.run_frr_command(commands)
        if result.success:
            log.info(f"Configured BGP AS {config.local_as} with router ID {config.router_id}")
            
        return result

    def configure_address_family(self, asn: int, networks: List[str], redistribute: List[Dict[str, Any]] = None) -> FRRCommandResult:
        """
        Configure BGP address family with specific networks and redistribution.
        
        Args:
            asn: BGP autonomous system number
            networks: List of networks to advertise (in CIDR notation)
            redistribute: List of redistribution configurations
            
        Returns:
            FRRCommandResult: Result of the operation
        """
        commands = [
            "enable",
            "configure terminal",
            f"router bgp {asn}",
            "address-family ipv4 unicast",
        ]
        
        # Add networks
        for network in networks:
            commands.append(f"network {network}")
            
        # Add redistributions
        if redistribute:
            for redist in redistribute:
                redist_cmd = f"redistribute {redist['protocol']}"
                if 'metric' in redist:
                    redist_cmd += f" metric {redist['metric']}"
                if 'route_map' in redist:
                    redist_cmd += f" route-map {redist['route_map']}"
                commands.append(redist_cmd)
                
        commands.append("exit-address-family")
        commands.append("exit")
        
        result = self.run_frr_command(commands)
        if result.success:
            log.info(f"Configured address-family ipv4 unicast for BGP AS {asn}")
            
        return result

    def configure_route_map(self, name: str, sequence: int, action: str, match_conditions: Dict[str, Any] = None, set_actions: Dict[str, Any] = None) -> FRRCommandResult:
        """
        Configure a route-map with match and set clauses.
        
        Args:
            name: Name of the route-map
            sequence: Sequence number
            action: 'permit' or 'deny'
            match_conditions: Dictionary of match conditions
            set_actions: Dictionary of set actions
            
        Returns:
            FRRCommandResult: Result of the operation
        """
        commands = [
            "enable",
            "configure terminal",
            f"route-map {name} {action} {sequence}",
        ]
        
        # Add match conditions
        if match_conditions:
            for condition, value in match_conditions.items():
                if isinstance(value, list):
                    for item in value:
                        commands.append(f"match {condition} {item}")
                else:
                    commands.append(f"match {condition} {value}")
                    
        # Add set actions
        if set_actions:
            for action_name, value in set_actions.items():
                if isinstance(value, list):
                    for item in value:
                        commands.append(f"set {action_name} {item}")
                else:
                    commands.append(f"set {action_name} {value}")
                    
        commands.append("exit")
        
        result = self.run_frr_command(commands)
        if result.success:
            log.info(f"Configured route-map {name} {action} {sequence}")
            
        return result

    def configure_access_list(self, name: str, entries: List[Dict[str, Any]]) -> FRRCommandResult:
        """
        Configure an access list with multiple entries.
        
        Args:
            name: Name or number of the access list
            entries: List of access list entries
            
        Returns:
            FRRCommandResult: Result of the operation
        """
        commands = [
            "enable",
            "configure terminal",
        ]
        
        for entry in entries:
            action = entry.get("action", "permit")
            prefix = entry.get("prefix", "")
            commands.append(f"access-list {name} {action} {prefix}")
            
        commands.append("exit")
        
        result = self.run_frr_command(commands)
        if result.success:
            log.info(f"Configured access-list {name} with {len(entries)} entries")
            
        return result

    def configure_prefix_list(self, name: str, entries: List[Dict[str, Any]]) -> FRRCommandResult:
        """
        Configure a prefix list with multiple entries.
        
        Args:
            name: Name of the prefix list
            entries: List of prefix list entries
            
        Returns:
            FRRCommandResult: Result of the operation
        """
        commands = [
            "enable",
            "configure terminal",
        ]
        
        for entry in entries:
            seq = entry.get("seq", "")
            action = entry.get("action", "permit")
            prefix = entry.get("prefix", "")
            le = entry.get("le", "")
            ge = entry.get("ge", "")
            
            cmd = f"ip prefix-list {name}"
            if seq:
                cmd += f" seq {seq}"
            cmd += f" {action} {prefix}"
            if le:
                cmd += f" le {le}"
            if ge:
                cmd += f" ge {ge}"
                
            commands.append(cmd)
            
        commands.append("exit")
        
        result = self.run_frr_command(commands)
        if result.success:
            log.info(f"Configured prefix-list {name} with {len(entries)} entries")
            
        return result

    def configure_static_route(self, prefix: str, next_hop: str) -> FRRCommandResult:
        """
        Configure a static route.
        
        Args:
            prefix: Destination prefix in CIDR notation
            next_hop: Next hop IP address or interface
            
        Returns:
            FRRCommandResult: Result of the operation
        """
        commands = [
            "enable",
            "configure terminal",
            f"ip route {prefix} {next_hop}",
            "exit",
        ]
        
        result = self.run_frr_command(commands)
        if result.success:
            log.info(f"Configured static route for {prefix} via {next_hop}")
            
        return result

    def configure_wireguard_interface(self, tunnel: CreatedWireguardTunnel, ospf_area: str = "0.0.0.0", ospf_process_id: int = 2) -> FRRCommandResult:
        """
        Configure a WireGuard tunnel interface.
        
        Args:
            tunnel: WireGuard tunnel configuration
            ospf_area: OSPF area ID
            ospf_process_id: OSPF process ID
            
        Returns:
            FRRCommandResult: Result of the operation
        """
        # First configure the basic interface
        result = self.configure_interface(
            tunnel.name,
            tunnel.ip_address,
            ospf_area=ospf_area,
            ospf_process_id=ospf_process_id,
            hello_interval=2,
            dead_interval=8,
            cost=10,
            network_type="point-to-point",
            description=f"WireGuard tunnel to {tunnel.peer_public_key[:8]}..."
        )
        
        return result

    def apply_legacy_config(self) -> bool:
        """
        Apply the legacy configuration for backward compatibility.
        
        Returns:
            bool: True if all configurations were applied successfully, False otherwise
        """
        success = True
        
        # Add loopback interface
        if self.route_reflector and hasattr(self.route_reflector, 'loopback'):
            result = self.add_loopback_interface(
                self.route_reflector.loopback + "/24", 
                ospf_area="0.0.0.0", 
                ospf_process_id=2
            )
            success = success and result.success
        
        # Add OSPF config
        if self.ospf_config:
            ospf_config = OspfConfig(
                process_id=2,
                router_id=self.ospf_config.router_id,
                areas=[OspfArea(area_id="0.0.0.0")],
                interfaces=[]
            )
            result = self.configure_ospf(ospf_config)
            success = success and result.success
        
        # Add address family
        if self.vpp_lan_interface:
            ip_parts = self.vpp_lan_interface.ip_address.split(".")
            ip_parts[-1] = "0"
            network = ".".join(ip_parts) + "/24"
            
            result = self.configure_address_family(
                64512, 
                [network],
                [{"protocol": "ospf", "ospf_id": 1}]
            )
            success = success and result.success
        
        # Add core BGP config
        if self.route_reflector and self.route_reflector_controller:
            commands = [
                "enable",
                "configure terminal",
                "router bgp 64512",
                f"bgp router-id {self.route_reflector.loopback}",
                f"neighbor {self.route_reflector_controller.loopback.split('/')[0]} remote-as internal",
                f"neighbor {self.route_reflector_controller.loopback.split('/')[0]} update-source lo",
                "neighbor " + self.route_reflector_controller.loopback.split("/")[0] + " timers connect 10",
                "exit"
            ]
            result = self.run_frr_command(commands)
            success = success and result.success
        
        # Add client OSPF config if enabled
        if self.ospf_config and getattr(self.ospf_config, 'client_enabled', False):
            if self.vpp_lan_interface:
                result = self.configure_interface(
                    self.vpp_lan_interface.interface_name,
                    f"{self.vpp_lan_interface.ip_address}/{self.vpp_lan_interface.prefix_len}",
                    ospf_area=self.ospf_config.client_router_area,
                    ospf_process_id=1
                )
                success = success and result.success
            
            commands = [
                "enable",
                "configure terminal",
                "router ospf 1",
                f"router-id {self.ospf_config.client_router_id}",
                "redistribute bgp",
                "exit"
            ]
            result = self.run_frr_command(commands)
            success = success and result.success
        
        # Configure WireGuard tunnels
        for wg in self.wireguard_tunnels:
            result = self.configure_wireguard_interface(wg)
            success = success and result.success
        
        if success:
            log.info("All legacy FRR configurations applied successfully!")
        else:
            log.error("Some legacy FRR configurations failed to apply.")
        
        return success

    def apply_config(self) -> bool:
        """
        Apply comprehensive FRR configuration.
        
        Returns:
            bool: True if all configurations were applied successfully, False otherwise
        """
        if not self.device_config:
            return self.apply_legacy_config()
        
        success = True
        
        # Configure interfaces
        for interface in self.interfaces:
            result = self.configure_interface(
                interface.name,
                f"{interface.ip_address}/{interface.subnet_mask.prefixlen}",
                description=interface.description
            )
            success = success and result.success
        
        # Configure OSPF
        if self.ospf_config:
            result = self.configure_ospf(self.ospf_config)
            success = success and result.success
            
            # Configure OSPF interfaces
            for ospf_iface in self.ospf_config.interfaces:
                # Find the corresponding interface
                for iface in self.interfaces:
                    if iface.name == ospf_iface.interface_name:
                        result = self.configure_interface(
                            ospf_iface.interface_name,
                            f"{iface.ip_address}/{iface.subnet_mask.prefixlen}",
                            ospf_area=ospf_iface.area_id,
                            ospf_process_id=self.ospf_config.process_id,
                            hello_interval=ospf_iface.hello_interval,
                            dead_interval=ospf_iface.dead_interval,
                            cost=ospf_iface.cost,
                            passive=ospf_iface.passive
                        )
                        success = success and result.success
                        break
        
        # Configure BGP
        if self.bgp_config:
            result = self.configure_bgp(self.bgp_config)
            success = success and result.success
        
        # Configure static routes
        if self.device_config and self.device_config.static_routes:
            for prefix, next_hop in self.device_config.static_routes.items():
                result = self.configure_static_route(str(prefix), str(next_hop))
                success = success and result.success
        
        # Configure route maps
        if self.device_config and self.device_config.route_maps:
            for route_map in self.device_config.route_maps:
                result = self.configure_route_map(
                    route_map.name,
                    route_map.sequence,
                    route_map.action,
                    route_map.match_conditions,
                    route_map.set_actions
                )
                success = success and result.success
        
        # Configure access lists
        if self.device_config and self.device_config.access_lists:
            for access_list in self.device_config.access_lists:
                result = self.configure_access_list(access_list.name, access_list.entries)
                success = success and result.success
        
        # Configure prefix lists
        if self.device_config and self.device_config.prefix_lists:
            for prefix_list in self.device_config.prefix_lists:
                result = self.configure_prefix_list(prefix_list.name, prefix_list.entries)
                success = success and result.success
        
        # Configure WireGuard tunnels
        for wg in self.wireguard_tunnels:
            result = self.configure_wireguard_interface(wg)
            success = success and result.success
        
        if success:
            log.info("All FRR configurations applied successfully!")
        else:
            log.error("Some FRR configurations failed to apply.")
        
        return success

    def clear_config(self) -> bool:
        """
        Clear all FRR configurations.
        
        Returns:
            bool: True if all configurations were cleared successfully, False otherwise
        """
        commands = [
            "enable",
            "configure terminal",
            "no router ospf 1",
            "no router ospf 2",
            "no router bgp 64512",
            "exit"
        ]
        
        result = self.run_frr_command(commands)
        if result.success:
            log.info("All FRR configurations cleared successfully!")
        else:
            log.error("Failed to clear FRR configurations.")
        
        return result.success



"""add_protocol_enablement_flags_to_tunnel_config

Revision ID: c8f9d2e5a1b3
Revises: b4d7e8f5c1a2
Create Date: 2025-07-09 15:00:00.000000

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'c8f9d2e5a1b3'
down_revision: Union[str, None] = 'b4d7e8f5c1a2'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Add ospf_enabled and ebgp_enabled boolean columns to tunnel_config_data table."""
    # Add ospf_enabled column
    op.add_column('tunnel_config_data', sa.Column(
        'ospf_enabled', 
        sa.Boolean(), 
        nullable=True,
        comment='Whether OSPF protocol is enabled for this configuration'
    ))
    
    # Add ebgp_enabled column
    op.add_column('tunnel_config_data', sa.Column(
        'ebgp_enabled', 
        sa.Boolean(), 
        nullable=True,
        comment='Whether eBGP protocol is enabled for this configuration'
    ))
    
    # Set default values for existing records
    op.execute("UPDATE tunnel_config_data SET ospf_enabled = false WHERE ospf_enabled IS NULL")
    op.execute("UPDATE tunnel_config_data SET ebgp_enabled = false WHERE ebgp_enabled IS NULL")


def downgrade() -> None:
    """Remove ospf_enabled and ebgp_enabled columns from tunnel_config_data table."""
    op.drop_column('tunnel_config_data', 'ebgp_enabled')
    op.drop_column('tunnel_config_data', 'ospf_enabled')



import os
import subprocess
from pprint import pprint
from typing import List, Tuple
from loguru import logger as log
from vpp_vrouter.client import ExtendedVPPAPIClient
from vpp_vrouter.common import models

from vrouter_agent.utils import run_command
from vrouter_agent.vpp.vpp_api import (
    add_gre_tunnel_to_wg_and_lcp_it_to_frr_container,
    add_wireguard_interface,
    add_wireguard_peer,
    add_wireguard_tunnel,
)
from vrouter_agent.vpp.vpp_utils import (
    WG_TUNNELS_PER_CONNECTION,
)
from vrouter_agent.utils.secure_subprocess import run_secure_command


def add_route_to_wireguard_peer(
    interface_name: str, remote_ip_network: str, remote_tunnel_end_ip: str
):
    run_command(
        [
            "sudo",
            "vppctl",
            "ip",
            "route",
            "add",
            remote_ip_network,
            "via",
            remote_tunnel_end_ip,
            interface_name,
        ]
    )


def add_frr_agent_ip_translation_from_gre_to_wg(
    gre_interface_ip_address: str,
    interface_ip: str,
    remote_gre_interface_ip_address: str,
    remote_tunnel_end_ip: str,
    frr_container_name: str = "frr",
    replace_file_path: str = "/rr-ip-replace.txt",
):
    # TODO when proper python client API will be created for this then replace this with python client code
    run_command(
        [
            "docker",
            "exec",
            "-t",
            frr_container_name,
            "sh",
            "-c",
            "echo '"
            + gre_interface_ip_address
            + "=>"
            + interface_ip
            + "' >> "
            + replace_file_path,
        ]
    )
    run_command(
        [
            "docker",
            "exec",
            "-t",
            frr_container_name,
            "sh",
            "-c",
            "echo '"
            + remote_gre_interface_ip_address
            + "=>"
            + remote_tunnel_end_ip
            + "' >> "
            + replace_file_path,
        ]
    )


def add_bgp_loop_interface_with_lcp():
    run_command(["sudo", "vppctl", "create", "loopback", "interface"])
    # run_command(["sudo", "vppctl", "lcp", "create", "loop0", "host-if", "bgp-loop", "netns", os.environ["NODE_FRR_NETNS"], "tun"])
    run_command(["sudo", "vppctl", "set", "int", "state", "loop0", "up"])
    run_command(
        [
            "sudo",
            "vppctl",
            "set",
            "int",
            "ip",
            "address",
            "loop0",
            os.environ["LAB_" + os.environ["NODE"] + "_BGP_LOOP_IP"] + "/32",
        ]
    )


def add_frr_configuration(
    wg_interface_info: List[Tuple[str, str, int]],
    use_ebgp: bool,
    customer_filtering: str = "",
    ospf_zebra_filtering: str = "",
):
    customer_filtering_route_map = (
        f"""{customer_filtering}
ip prefix-list other seq 5 permit any
!
route-map customer-filtering deny 5
 match ip address prefix-list customer-underlay
!
route-map customer-filtering permit 10
 match ip address prefix-list other
!"""
        if customer_filtering
        else ""
    )
    ospf_zebra_route_map = (
        f"""
!
! filtering out installation of WG and BGP loopback subnets that are not needed for this node
ip protocol ospf route-map ospf-zebra-filtering
!
{ospf_zebra_filtering}
ip prefix-list other2 seq 5 permit any
!
route-map ospf-zebra-filtering permit 5
match ip address prefix-list wg-and-bgp-loopbacks-needed
!
route-map ospf-zebra-filtering deny 10
 match ip address prefix-list all-wg-and-bgp-loopbacks
!
route-map ospf-zebra-filtering permit 15
 match ip address prefix-list other2
!"""
        if ospf_zebra_filtering
        else ""
    )

    wg_interface_frr_config = ""
    for wg_name, wg_ip, wg_link_index in wg_interface_info:
        wg_interface_frr_config += f"""!
interface {wg_name}
 ip address {wg_ip}
 # 0.0.0.0 is static as all core-OSPF ends on all nodes are in the same OSPF area
 ip ospf 2 area 0.0.0.0
 ip ospf network point-to-point
 ip ospf hello-interval {os.environ["VPP_CORE_OSPF_HELLO_INTERVAL"]}
 ip ospf dead-interval {os.environ["VPP_CORE_OSPF_DEAD_INTERVAL"]}
 ip ospf cost {os.environ[f"VPP_CORE_OSPF_NODE_CONNECTION{wg_link_index}_COST"]}
!
"""
    if use_ebgp:
        customer_filtering_for_neighbor = (
            f"neighbor {os.environ[os.environ['CLIENT'] + '_BGP_LOOP_IP']} route-map customer-filtering out"
            if customer_filtering
            else "!"
        )
        customer_interface_config = f"""interface {os.environ[os.environ["VPP"] + "_LAN_INTF"]}
 ip address {os.environ[os.environ["VPP"] + "_LAN_INTF_IP"]}/{os.environ[os.environ["VPP"] + "_LAN_INTF_IP_MASK"]}
!"""
        ospf_one_config = "!"
        bgp_router_and_filtering_config = f"""!{customer_filtering_route_map}
# BGP ASN is the same for all BGP nodes -> 64512 is static ASN
router bgp 64512
 bgp router-id {os.environ["LAB_" + os.environ["NODE"] + "_BGP_LOOP_IP"]}
 # need to disable default need for policy to be talk to other eBGP peers
 no bgp ebgp-requires-policy
# no bgp suppress-duplicates
# no bgp network import-check
 neighbor {os.environ["LAB_RR_BGP_LOOP_IP"]} remote-as internal
 neighbor {os.environ["LAB_RR_BGP_LOOP_IP"]} update-source lo
 # change next hop to self to not need to distribute also customer's eBGP router id (bgp loop ip)
 neighbor {os.environ["LAB_RR_BGP_LOOP_IP"]} next-hop-self
 # in case of peer connection failure, retry connection every 10s
 neighbor {os.environ["LAB_RR_BGP_LOOP_IP"]} timers connect {os.environ["BGP_PEER_CONNECTION_RETRY_INTERVAL"]}
# neighbor {os.environ["LAB_RR_BGP_LOOP_IP"]} timers 5 15
 neighbor {os.environ[os.environ["CLIENT"] + "_BGP_LOOP_IP"]} remote-as external
 neighbor {os.environ[os.environ["CLIENT"] + "_BGP_LOOP_IP"]} update-source lo
 neighbor {os.environ[os.environ["CLIENT"] + "_BGP_LOOP_IP"]} ebgp-multihop 20
 {customer_filtering_for_neighbor}
 !
 address-family ipv4 unicast
  network {os.environ[os.environ["VPP"] + "_LAN_INTF_IP_NETWORK"]}
 exit-address-family
!{ospf_zebra_route_map}"""

    else:  # customer OSPF (OSPF1)
        customer_filtering_route_map_ref = (
            " route-map customer-filtering" if customer_filtering else ""
        )
        customer_interface_config = f"""interface {os.environ[os.environ["VPP"] + "_LAN_INTF"]}
 ip address {os.environ[os.environ["VPP"] + "_LAN_INTF_IP"]}/{os.environ[os.environ["VPP"] + "_LAN_INTF_IP_MASK"]}
 ip ospf 1 area {os.environ[os.environ["VPP"] + "_CUSTOMER_OSPF_AREA"]}
 ip ospf network point-to-point
 ip ospf hello-interval {os.environ["VPP_CUSTOMER_OSPF_HELLO_INTERVAL"]}
 ip ospf dead-interval {os.environ["VPP_CUSTOMER_OSPF_DEAD_INTERVAL"]}
!"""
        ospf_one_config = f"""router ospf 1
 ospf router-id {os.environ[os.environ["VPP"] + "_CUSTOMER_OSPF_ROUTER_ID"]}
 redistribute bgp{customer_filtering_route_map_ref}
!"""
        bgp_router_and_filtering_config = f"""# BGP ASN is the same for all BGP nodes -> 64512 is static ASN
router bgp 64512
 bgp router-id {os.environ["LAB_" + os.environ["NODE"] + "_BGP_LOOP_IP"]}
# no bgp ebgp-requires-policy
# no bgp suppress-duplicates
# no bgp network import-check
 neighbor {os.environ["LAB_RR_BGP_LOOP_IP"]} remote-as internal
 neighbor {os.environ["LAB_RR_BGP_LOOP_IP"]} update-source lo
 # in case of peer connection failure, retry connection every 10s
 neighbor {os.environ["LAB_RR_BGP_LOOP_IP"]} timers connect {os.environ["BGP_PEER_CONNECTION_RETRY_INTERVAL"]}
# neighbor {os.environ["LAB_RR_BGP_LOOP_IP"]} timers 5 15
 !
 address-family ipv4 unicast
  network {os.environ[os.environ["VPP"] + "_LAN_INTF_IP_NETWORK"]}
  redistribute ospf 1
 exit-address-family
!{customer_filtering_route_map}{ospf_zebra_route_map}"""

    config_str = f"""!
frr version 8.3
frr defaults traditional
hostname bgp{os.environ[os.environ["NODE"] + "_INDEX"]}
service integrated-vtysh-config
log file /var/log/frr/debug.log
log stdout
log commands
no ipv6 forwarding
!
debug zebra events
debug bgp zebra
debug bgp neighbor-events
debug bgp updates
debug bgp keepalives
debug bgp update-groups
debug bgp nht
debug ospf event
debug ospf zebra
#debug ospf packet all send detail
#debug ospf packet all recv detail
debug static events
debug static route
!
interface lo
 ip address {os.environ["LAB_" + os.environ["NODE"] + "_BGP_LOOP_IP"]}/32
 ip ospf 2 area 0.0.0.0
!
{wg_interface_frr_config}
{customer_interface_config}
{ospf_one_config}
router ospf 2
 ospf router-id {os.environ[os.environ["VPP"] + "_CORE_OSPF_ROUTER_ID"]}
!
{bgp_router_and_filtering_config}
line vty
!
"""
    # clean previous configuration (the vtysh -b just adds stuff, but can delete it -> need to wipe out everything)
    # subprocess.run(
    #     "sudo sh -c \"echo '' > /etc/frr/frr.conf\"", shell=True, capture_output=True
    # )
    # subprocess.run("sudo vtysh -b", shell=True, capture_output=True)
    # time.sleep(2)

    # add new configuration
    with open(os.environ["NODE_FRR_TMP_FRR_CONFIG_FILE"], "w") as frr_config:
        frr_config.write(config_str)
    return True


def add_nat_configuration(
    wg_ports_to_open: List[int], tcp_ports: List[int] = [], udp_ports: List[int] = []
):
    run_command(
        [
            "sudo",
            "vppctl",
            "nat44",
            "plugin",
            "enable",
            "sessions",
            os.environ["VPP_NAT_SESSION_COUNT"],
        ]
    )
    run_command(
        [
            "sudo",
            "vppctl",
            "set",
            "int",
            "nat44",
            "in",
            os.environ[os.environ["VPP"] + "_LAN_INTF"],
        ]
    )
    run_command(
        [
            "sudo",
            "vppctl",
            "set",
            "int",
            "nat44",
            "out",
            os.environ[os.environ["VPP"] + "_WAN_INTF"],
        ]
    )
    run_command(
        [
            "sudo",
            "vppctl",
            "nat44",
            "add",
            "address",
            os.environ[os.environ["VPP"] + "_WAN_INTF_IP"],
        ]
    )  # ip address of eth4
    # run_command(["sudo", "vppctl", "nat44", "forward", "enable"])  # disables the NAT drop of unknown UDP, but it doesn't help the performance :/
    for port in wg_ports_to_open:
        run_command(
            [
                "sudo",
                "vppctl",
                "nat44",
                "add",
                "identity",
                "mapping",
                "external",
                os.environ[os.environ["VPP"] + "_WAN_INTF"],
                "udp",  # wireguard's encrypted communication are UDP packets
                str(port),
            ]
        )
    for port in tcp_ports:
        run_command(
            [
                "sudo",
                "vppctl",
                "nat44",
                "add",
                "identity",
                "mapping",
                "external",
                os.environ[os.environ["VPP"] + "_WAN_INTF"],
                "tcp",
                str(port),
            ]
        )
    for port in udp_ports:
        run_command(
            [
                "sudo",
                "vppctl",
                "nat44",
                "add",
                "identity",
                "mapping",
                "external",
                os.environ[os.environ["VPP"] + "_WAN_INTF"],
                "udp",
                str(port),
            ]
        )


def add_acl_configuration():
    # NOTE: this is test ACL configuration that has nothing to do with productions requirements (unknown currently)
    # -> will parametrize into env file only the one ip address that is denied
    run_command(
        [
            "sudo",
            "vppctl",
            "set",
            "acl-plugin",
            "acl",
            "deny",
            "dst",
            os.environ["VPP_ACL_DENIED_IP"] + "/32,permit",
            "tag",
            "my-test-rules",
        ]
    )
    run_command(
        [
            "sudo",
            "vppctl",
            "set",
            "acl-plugin",
            "interface",
            os.environ[os.environ["VPP"] + "_LAN_INTF"],
            "input",
            "acl",
            "0",
        ]
    )
    # for ACL info use this VPP CLI:
    # sh acl-plugin acl
    # sh acl-plugin interface sw_if_index 1


def add_client_neighbour_info():
    """Adds neighbor L2 information to VPP so that VPP won't drop packets from TRex and send ARP packets(=question
    for client)"""

    # using MAC address of client interface (or at least what TREX traffic destination MAC address that VPP must
    # steer back to Trex through client interface)
    run_command(
        [
            "sudo",
            "vppctl",
            "ip",
            "neighbor",
            os.environ[os.environ["VPP"] + "_LAN_INTF"],
            os.environ[os.environ["CLIENT"] + "_VPP_INTF_IP"],
            os.environ[os.environ["CLIENT"] + "_VPP_INTF_MAC"],
        ]
    )
    # NOTE: this could be also moved to bootstrap.vpp as it is static for given vpp-client setup


# Note: this does not include RR-node wireguard tunnel (for that see makefile) and RR setup => only path from linux WG
# on node to BGP in container on node
def add_routes_for_bidirectional_bgp_communication():
    # Some path explanation:
    # RR-to-node path: node-linux-wg0 -(adding here route1+iptables forwarding)-> FRR container default interface ->
    #   FRR container routing to local loopback -(ping echo going back, need to add route2)-> FRR container default
    #   interface -(routing path was added as part of wg configuration)-> node-linux-wg0
    # Node-to-RR path: FRR instance in FRR container -(need to add route3)-> FRR container default interface
    #   -> node linux host route(actually controlled by WG allowed ip configuration) -> wg tunnel to RR

    # this can be hardcoded unless you want change default networking for all docker containers
    host_ip_from_docker_container = "172.17.0.1"

    # adding route1
    frr_container_ip_address = run_command(
        [
            "docker",
            "inspect",
            "-f",
            "'{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}'",
            os.environ["NODE_FRR_CONTAINER_NAME"],
        ]
    )
    frr_container_ip_address = frr_container_ip_address.replace("\n", "").replace(
        "'", ""
    )
    try:
        run_command(
            [
                "sudo",
                "ip",
                "route",
                "add",
                os.environ["LAB_" + os.environ["NODE"] + "_BGP_LOOP_IP"],
                "via",
                frr_container_ip_address,
            ]
        )
    except RuntimeError:
        pass  # route already exists
    # allow packet pass from safe wireguard interface into docker container (i guess it is disabled by default because
    # safe wireguard interface is external interface and therefore seen as unsafe outside world)
    run_command(
        [
            "sudo",
            "iptables",
            "-A",
            "FORWARD",
            "-i",
            os.environ["LAB_" + os.environ["NODE"] + "_LINUX_WG_INTERFACE"],
            "-o",
            "docker0",  # static for default docker installation
            "-j",
            "ACCEPT",
        ]
    )
    # adding route2
    try:
        run_command(
            [
                "docker",
                "exec",
                "-t",
                os.environ["NODE_FRR_CONTAINER_NAME"],
                "ip",
                "route",
                "add",
                os.environ["LAB_RR_WG_INTERFACE_IP"],
                "via",
                host_ip_from_docker_container,
            ]
        )
    except RuntimeError:
        pass

    try:
        # adding route3
        run_command(
            [
                "docker",
                "exec",
                "-t",
                os.environ["NODE_FRR_CONTAINER_NAME"],
                "ip",
                "route",
                "add",
                os.environ["LAB_RR_BGP_LOOP_IP"],
                "via",
                host_ip_from_docker_container,
            ]
        )
    except RuntimeError:
        pass

    # TODO check that setting "echo 1 > /proc/sys/net/ipv4/ip_forward" is always set for node (it seems to be true,
    #  just don't know where it is persisted to survive the machine reboot)


def add_ospf_lcp_fix():
    # https://lists.fd.io/g/vpp-dev/topic/83103366#19478
    # for 224.0.0.5 and 224.0.0.6 used by OSPF (https://en.wikipedia.org/wiki/Open_Shortest_Path_First)

    # run_command(["sudo", "vppctl", "ip", "mroute", "add", "224.0.0.0/24", "via", "local", "Forward"])
    # for i in range(4):
    #     run_command(["sudo", "vppctl", "ip", "mroute", "add", "224.0.0.0/24", "via", os.environ[os.environ["VPP"]+f"_WG{i}_INTERFACE_NAME"], "Accept"])

    # to steer OSPF Hello packets from FRR container to client
    run_command(
        [
            "sudo",
            "vppctl",
            "ip",
            "route",
            "add",
            "224.0.0.0/24",  # static due to OSPF hello packets always destined for 224.0.0.5
            "via",
            os.environ[os.environ["CLIENT"] + "_VPP_INTF_IP"],
            os.environ[os.environ["VPP"] + "_LAN_INTF"],
        ]
    )

    # to get ospf hello packets from client to frr container through vpp
    run_command(
        [
            "sudo",
            "vppctl",
            "ip",
            "mroute",
            "add",
            os.environ[os.environ["CLIENT"] + "_VPP_INTF_IP"],
            "224.0.0.5",  # static destination ip address of Hello OSPF packets, fixed by OSPF protocol itself
            "via",
            os.environ[os.environ["VPP"] + "_LAN_INTF"],
            "Accept",
        ]
    )
    run_command(
        [
            "sudo",
            "vppctl",
            "ip",
            "mroute",
            "add",
            os.environ[os.environ["CLIENT"] + "_VPP_INTF_IP"],
            "224.0.0.5",  # static destination ip address of Hello OSPF packets, fixed by OSPF protocol itself
            "via",
            "local",  # static, set to be locally processed (processed by LCP)
            "Forward",
        ]
    )


def add_lcp_global_configuration():
    run_command(
        ["sudo", "vppctl", "lcp", "default", "netns", os.environ["NODE_FRR_NETNS"]]
    )
    run_command(["sudo", "vppctl", "lcp", "lcp-sync", "on"])  # static LCP setting
    run_command(
        ["sudo", "vppctl", "lcp", "lcp-auto-subint", "on"]
    )  # static LCP setting


def add_tap_tunnel_to_dhcp_container():
    run_command(
        [
            "sudo",
            "vppctl",
            "create",
            "tap",
            "id",
            os.environ["NODE_DHCP_VPP_INTF_TAP_ID"],
            "host-ns",
            os.environ["NODE_DHCP_NETNS"],
            "host-ip4-addr",
            os.environ["NODE_DHCP_CONTAINER_INTF_IP"]
            + "/"
            + os.environ["NODE_DHCP_CONTAINER_INTF_IP_MASK"],
            "host-if-name",
            os.environ["NODE_DHCP_CONTAINER_INTF"],
            "host-mtu-size",
            os.environ["NODE_DHCP_CONTAINER_INTF_MTU"],
        ]
    )
    run_command(
        [
            "sudo",
            "vppctl",
            "set",
            "interface",
            "state",
            os.environ["NODE_DHCP_VPP_INTF"],
            "up",
        ]
    )
    run_command(
        [
            "sudo",
            "vppctl",
            "set",
            "interface",
            "ip",
            "address",
            os.environ["NODE_DHCP_VPP_INTF"],
            os.environ["NODE_DHCP_VPP_INTF_IP"]
            + "/"
            + os.environ["NODE_DHCP_VPP_INTF_IP_MASK"],
        ]
    )


def add_dhcp_broadcast_forwarding():
    run_command(
        [
            "sudo",
            "vppctl",
            "set",
            "acl-plugin",
            "acl",
            "permit",
            "dst",
            "255.255.255.255/32",  # broadcast address is constant
            "dport",
            "67",  # static DHCP port defined in DHCP protocol (port on client side)
        ]
    )
    run_command(
        [
            "sudo",
            "vppctl",
            "abf",
            "policy",
            "add",
            "id",
            os.environ["NODE_DHCP_ABF_INDEX_FOR_PATH_TO_DHCP_SERVER"],
            "acl",
            "1",
            # this index should be index of ACL created by previous call -> hardcoded for now # FIXME pythonization should use ACL index from previous call
            "via",
            os.environ["NODE_DHCP_CONTAINER_INTF_IP"],
            os.environ["NODE_DHCP_VPP_INTF"],
        ]
    )
    run_command(
        [
            "sudo",
            "vppctl",
            "abf",
            "attach",
            "ip4",
            "policy",
            os.environ["NODE_DHCP_ABF_INDEX_FOR_PATH_TO_DHCP_SERVER"],
            os.environ[os.environ["VPP"] + "_LAN_INTF"],
        ]
    )

    run_command(
        [
            "sudo",
            "vppctl",
            "set",
            "acl-plugin",
            "acl",
            "permit",
            "dst",
            "255.255.255.255/32",  # broadcast address is constant
            "dport",
            "68",  # another static DHCP port defined in DHCP protocol (port on DHCP server side)
        ]
    )
    run_command(
        [
            "sudo",
            "vppctl",
            "abf",
            "policy",
            "add",
            "id",
            os.environ["NODE_DHCP_ABF_INDEX_FOR_PATH_TO_DHCP_CLIENT"],
            "acl",
            "2",
            # this index should be index of ACL created by previous call -> hardcoded for now # FIXME pythonization should use ACL index from previous call
            "via",
            os.environ[os.environ["CLIENT"] + "_VPP_INTF_IP"],
            os.environ[os.environ["VPP"] + "_LAN_INTF"],
        ]
    )
    run_command(
        [
            "sudo",
            "vppctl",
            "abf",
            "attach",
            "ip4",
            "policy",
            os.environ["NODE_DHCP_ABF_INDEX_FOR_PATH_TO_DHCP_CLIENT"],
            os.environ["NODE_DHCP_VPP_INTF"],
        ]
    )


def add_dhcp_unicast_forwarding():
    # unicast should be handled by automatically added routes from ip address setting on interfaces (LAN and tap to
    # DHCP container), but arp neighbor checking can get into way at DHCP server to DHCP client path
    run_command(
        [
            "sudo",
            "vppctl",
            "ip",
            "neighbor",
            os.environ[os.environ["VPP"] + "_LAN_INTF"],
            os.environ[os.environ["CLIENT"] + "_VPP_INTF_IP"],
            os.environ[os.environ["CLIENT"] + "_VPP_INTF_MAC"],
        ]
    )


def add_dhcp_configuration():
    add_tap_tunnel_to_dhcp_container()
    add_dhcp_broadcast_forwarding()
    add_dhcp_unicast_forwarding()
    with ExtendedVPPAPIClient() as client:
        configure_dhcp(client)


def configure_dhcp(client):
    config = (
        """
{
    "Dhcp4": {
        // for all keys for this config check https://gitlab.isc.org/isc-projects/kea/-/blob/master/doc/examples/kea4/all-keys.json?ref_type=heads
        "subnet4": [
            {
                "subnet": """
        + '"'
        + os.environ[os.environ["CLIENT"] + "_VPP_INTF_IP_NETWORK"]
        + '"'
        + """,
                "pools": [
                    {
                        "pool": """
        + '"'
        + os.environ[os.environ["NODE"] + "_DHCP_START_IP"]
        + "-"
        + os.environ[os.environ["NODE"] + "_DHCP_END_IP"]
        + '"'
        + """
                    }
                ],
                // Specifies that this subnet is selected for requests
                // received on a particular interface.
                "interface": """
        + '"'
        + os.environ["NODE_DHCP_CONTAINER_INTF"]
        + '"'
        + """,

                // Specify whether the server should look up global reservations.
                "reservations-global": false,

                // Specify whether the server should look up in-subnet reservations.
                "reservations-in-subnet": true,

                // Specify whether the server can assume that all reserved addresses
                // are out-of-pool.
                // Ignored when reservations-in-subnet is false.
                // If specified, it is inherited by "shared-networks" and
                // "subnet4" levels.
                "reservations-out-of-pool": false,

                "reservations": [
                    // This is a reservation for a specific hardware/MAC address.
                    {
                        "hw-address": """
        + '"'
        + os.environ[os.environ["CLIENT"] + "_VPP_INTF_MAC"]
        + '"'
        + """,
                        "ip-address": """
        + '"'
        + os.environ[os.environ["CLIENT"] + "_VPP_INTF_IP"]
        + '"'
        + """,
                        "hostname": """
        + '"'
        + os.environ[os.environ["NODE"] + "_DHCP_HOST_NAME"]
        + '"'
        + """,

                        // Reservation-specific list of DHCP options.
                        // for all options check https://gitlab.isc.org/isc-projects/kea/-/blob/master/doc/examples/kea4/all-options.json?ref_type=heads
                        // or check options in DHCP rfc(could have different names but is the source of truth for DHCP protocol) https://datatracker.ietf.org/doc/html/rfc2132 (i.e https://datatracker.ietf.org/doc/html/rfc2132#section-3.5)
                        "option-data": [
                           {
                                "code": 3,
                                "data": """
        + '"'
        + os.environ[os.environ["VPP"] + "_LAN_INTF_IP"]
        + '"'
        + """,
                                // only 1 router, the default gateway
                                "name": "routers"
                           },
                           {
                                "code": 6,
                                "data": """
        + '"'
        + os.environ[os.environ["NODE"] + "_DHCP_DNS"]
        + '"'
        + """,
                                "name": "domain-name-servers"
                           },
                           {
                                "code": 15,
                                "data": """
        + '"'
        + os.environ[os.environ["NODE"] + "_DHCP_DOMAIN_NAME"]
        + '"'
        + """,
                                "name": "domain-name"
                           },
                           {
                                "code": 26,
                                "data": """
        + '"'
        + os.environ[os.environ["CLIENT"] + "_VPP_INTF_MTU"]
        + '"'
        + """,
                                "name": "interface-mtu"
                           },
                           {
                                "code": 42,
                                // ping of ntp.ubuntu.com
                                "data": """
        + '"'
        + os.environ[os.environ["NODE"] + "_DHCP_NTP_SERVERS"]
        + '"'
        + """,
                                "name": "ntp-servers"
                           },
                           {
                              "code": 121,
                              // please mind the convenience notation used:
                              // subnet1 - router1 IP addr, subnet2 - router2 IP addr, ..., subnetN - routerN IP addr
                              "data": """
        + '"'
        + os.environ["NODE_DHCP_CONTAINER_INTF_IP"]
        + "/32"
        + " - "
        + os.environ[os.environ["VPP"] + "_LAN_INTF_IP"]
        + '"'
        + """,
                              "name": "classless-static-route",
                              // needed to parse "data" as string and requiring "data" in hexadecimal format
                              "csv-format": true
                           }
                        ]
                    }
                ]
            }
        ],
        "interfaces-config": {
            // Specifies a list of interfaces on which the Kea DHCPv4
            // server should listen to DHCP requests.
            "interfaces": [
                """
        + '"'
        + os.environ["NODE_DHCP_CONTAINER_INTF"]
        + '"'
        + """
            ],
            "dhcp-socket-type": "raw",
            "service-sockets-max-retries": 20,
            "service-sockets-require-all": true
        },
        "control-socket": {
            "socket-type": "unix",
            "socket-name": """
        + '"'
        + os.environ["NODE_DHCP_CONTROL_SOCKET"]
        + '"'
        + """
        },
        // for timers meaning see http://www.tcpipguide.com/free/t_DHCPLeaseRenewalandRebindingProcesses-2.htm
        "renew-timer": """
        + os.environ["NODE_DHCP_RENEW_TIMER"]
        + """,
        "rebind-timer": """
        + os.environ["NODE_DHCP_REBIND_TIMER"]
        + """,
        "valid-lifetime": """
        + os.environ["NODE_DHCP_VALID_LIFETIME"]
        + """,
        "loggers": [
            {
                "name": "kea-dhcp4",
                "output-options": [
                    {
                        "output": "stdout"
                    }
                ],
                // "severity": "INFO"
                "severity": """
        + '"'
        + os.environ["NODE_DHCP_LOG_LEVEL"]
        + '"'
        + """,
                // Debug level, a value between 0..99. The greater the value
                // the more detailed the debug log.
                "debuglevel": """
        + os.environ["NODE_DHCP_DEBUG_LEVEL"]
        + """
            }
        ],
        "lease-database": {
            "type": "memfile",
            "persist": true,
            "name": """
        + '"'
        + os.environ["NODE_DHCP_LEASE_DB_FILE"]
        + '"'
        + """
        }
    }
}
    """
    )

    reply = client.add_configuration(
        models.DHCPConfigurationItem(
            config=config,
            restart_before_config_apply=True,  # need only when new interfaces were created after DHCP server start
        )
    )
    pprint(reply, width=200)  # TODO error handling


def add_multiple_wireguard_tunnels(
    client, vpp_pairs_to_connect_together, redundant_link_count=1
):
    for _, (src_vpp, dst_vpp) in enumerate(vpp_pairs_to_connect_together):
        for link_index in range(redundant_link_count):
            for i in range(WG_TUNNELS_PER_CONNECTION):
                wg_index = link_index * WG_TUNNELS_PER_CONNECTION + i
                wan_suffix = "2" if link_index == 1 else ""
                add_wireguard_tunnel(
                    client,
                    os.environ[
                        f"VPP{src_vpp}_TO_VPP{dst_vpp}_WG{wg_index}_INTERFACE_NAME"
                    ],
                    os.environ[
                        f"VPP{src_vpp}_TO_VPP{dst_vpp}_WG{wg_index}_INTERFACE_IP"
                    ],
                    os.environ[
                        f"VPP{src_vpp}_TO_VPP{dst_vpp}_WG{wg_index}_INTERFACE_IP"
                    ]
                    + "/"
                    + os.environ[
                        f"VPP{src_vpp}_TO_VPP{dst_vpp}_WG{wg_index}_INTERFACE_IP_MASK"
                    ],
                    int(os.environ[f"VPP{src_vpp}_TO_VPP{dst_vpp}_WG{wg_index}_PORT"]),
                    os.environ[
                        f"VPP{src_vpp}_TO_VPP{dst_vpp}_WG{wg_index}_PRIVATE_KEY"
                    ],
                    os.environ[f"VPP{dst_vpp}_TO_VPP{src_vpp}_WG{wg_index}_PUBLIC_KEY"],
                    os.environ[
                        f"VPP{dst_vpp}_TO_VPP{src_vpp}_WG{wg_index}_INTERFACE_IP_NETWORK"
                    ],
                    os.environ[
                        f"VPP{dst_vpp}_TO_VPP{src_vpp}_WG{wg_index}_INTERFACE_IP"
                    ],
                    os.environ[
                        f"VPP{src_vpp}_TO_VPP{dst_vpp}_WG{wg_index}_GRE_INTERFACE_NAME"
                    ],
                    os.environ[
                        f"VPP{src_vpp}_TO_VPP{dst_vpp}_WG{wg_index}_GRE_INTERFACE_IP_ADDRESS"
                    ],
                    os.environ[
                        f"VPP{src_vpp}_TO_VPP{dst_vpp}_WG{wg_index}_GRE_INTERFACE_IP_ADDRESS_MASK"
                    ],
                    os.environ[
                        f"VPP{dst_vpp}_TO_VPP{src_vpp}_WG{wg_index}_GRE_INTERFACE_IP_ADDRESS"
                    ],
                    os.environ[f"VPP{src_vpp}_WAN{wan_suffix}_INTF_IP"],
                    os.environ[f"VPP{dst_vpp}_WAN{wan_suffix}_INTF_IP"],
                    int(os.environ["VPP_WG_INTF_MTU"]),
                    int(os.environ["VPP_WG_PERSISTENT_KEEPALIVE"]),
                )


def apply_frr_config(self):
    try:
        # Use secure subprocess wrapper with proper command list
        cmd = ["docker", "exec", "frr", "cp", os.environ['NODE_FRR_TMP_FRR_CONFIG_FILE'], "/etc/frr/frr.conf"]
        stdout, stderr = run_secure_command(cmd)
        
        # Use secure subprocess wrapper for vtysh command
        vtysh_cmd = ["docker", "exec", "frr", "vtysh", "-b"]
        stdout, stderr = run_secure_command(vtysh_cmd)
        
        return True
    except Exception as e:
        log.error(f"Failed to apply FRR config: {e}")
        return False



"""
ACL (Access Control List) Configuration Manager

This module provides specialized management for ACL configurations,
implementing security management interfaces for network access control.
"""

from typing import List, Dict, Any, Optional
import ipaddress

from loguru import logger as log
from vpp_vrouter.common import models

from ..base_interfaces import SecurityManager, create_result_dict
from ..connection.vpp_connection import VPPConnectionManager
from ...core.base import Interface


class ACLManager(SecurityManager):
    """
    Manages ACL (Access Control List) configuration through VPP API.
    
    This class handles ACL rule creation, application to interfaces,
    and management of network access control policies.
    """
    
    def __init__(self, connection_manager: VPPConnectionManager):
        """
        Initialize the ACL manager.
        
        Args:
            connection_manager: VPP connection manager instance
        """
        self.connection = connection_manager
        self.active_policies: Dict[str, Dict[str, Any]] = {}
    
    def _extract_reply_errors(self, reply) -> List[str]:
        """Extract all error messages from a VPP reply object."""
        errors = []
        
        # Check processing_error (legacy field)
        if hasattr(reply, 'processing_error') and reply.processing_error:
            errors.append(reply.processing_error)
        
        # Check added_items for errors
        if hasattr(reply, 'added_items'):
            for item in reply.added_items:
                if hasattr(item, 'error') and item.error:
                    errors.append(f"Added item error: {item.error}")
                if hasattr(item, 'state') and 'FAILURE' in str(item.state):
                    errors.append(f"Item state failure: {item.state}")
        
        # Check vpp_apply_attempted_items for detailed errors
        if hasattr(reply, 'vpp_apply_attempted_items'):
            for item in reply.vpp_apply_attempted_items:
                if hasattr(item, 'error') and item.error:
                    errors.append(f"VPP apply error: {item.error}")
                if hasattr(item, 'state') and 'FAILURE' in str(item.state):
                    errors.append(f"VPP apply state failure: {item.state}")
        
        # Check vpp_apply_success flag
        if hasattr(reply, 'vpp_apply_success') and reply.vpp_apply_success is False:
            if not errors:  # Only add generic message if no specific errors found
                errors.append("VPP apply failed without specific error message")
        
        return errors

    def configure_acl(self, ingress_interface: Interface, deny_network: str,
                     additional_rules: List[Dict[str, Any]] = None) -> bool:
        """
        Configure ACL with basic deny rule and optional additional rules.
        
        Args:
            ingress_interface: Interface to apply ACL to
            deny_network: Network to deny (CIDR notation)
            additional_rules: Optional list of additional ACL rules
            
        Returns:
            bool: True if configuration successful, False otherwise
        """
        try:
            policy_name = f"acl-{ingress_interface.interface_name}"
            additional_rules = additional_rules or []
            
            log.info(f"Configuring ACL policy '{policy_name}' on {ingress_interface.interface_name}")
            
            # Create ACL rules
            rules = []
            
            # Add deny rule for specified network
            rules.append(
                models.ACLRuleConfigurationItem(
                    action=models.ACLAction.DENY,
                    refinement=models.IPSpecification(
                        addresses=models.IPAddresses(
                            destination_network=ipaddress.IPv4Network(deny_network)
                        )
                    ),
                )
            )
            
            # Add additional custom rules
            for rule_config in additional_rules:
                rule = self._create_acl_rule_from_config(rule_config)
                if rule:
                    rules.append(rule)
            
            # Add final permit-all rule (standard practice)
            rules.append(
                models.ACLRuleConfigurationItem(
                    action=models.ACLAction.PERMIT,
                )
            )
            
            # Create ACL configuration
            acl_config = models.ACLConfigurationItem(
                name=policy_name,
                ingress=[ingress_interface.interface_name],
                rules=rules,
            )
            
            # Apply configuration to VPP
            reply = self.connection.client.add_configuration(acl_config)
            
            errors = self._extract_reply_errors(reply)
            if errors:
                for error in errors:
                    log.error(f"Error adding ACL configuration: {error}")
                return False
            
            # Store policy information
            self.active_policies[policy_name] = {
                'interface': ingress_interface.interface_name,
                'deny_network': deny_network,
                'additional_rules': additional_rules,
                'rule_count': len(rules),
                'created_at': self._get_current_timestamp()
            }
            
            log.info(f"ACL policy '{policy_name}' applied successfully with {len(rules)} rules")
            return True
            
        except Exception as e:
            log.error(f"Exception configuring ACL: {e}")
            return False
    
    def remove_acl_configuration(self, interface: Interface) -> bool:
        """
        Remove ACL configuration from an interface.
        
        Args:
            interface: Interface to remove ACL from
            
        Returns:
            bool: True if removal successful, False otherwise
        """
        try:
            policy_name = f"acl-{interface.interface_name}"
            
            log.info(f"Removing ACL configuration '{policy_name}'")
            
            # Find existing ACL configuration
            current_config = self.connection.client.get_configuration()
            
            acl_configs = [
                item.config
                for item in current_config.items
                if isinstance(item.config, models.ACLConfigurationItem)
                and item.config.ingress == [interface.interface_name]
            ]
            
            if not acl_configs:
                log.warning(f"No ACL configuration found for interface {interface.interface_name}")
                return True
            
            # Remove configurations from VPP
            reply = self.connection.client.delete_configuration(*acl_configs)
            
            errors = self._extract_reply_errors(reply)
            if errors:
                for error in errors:
                    log.error(f"Error deleting ACL configuration: {error}")
                return False
            
            # Remove from active policies
            if policy_name in self.active_policies:
                del self.active_policies[policy_name]
            
            log.info(f"ACL configuration '{policy_name}' removed successfully")
            return True
            
        except Exception as e:
            log.error(f"Exception removing ACL configuration: {e}")
            return False
    
    def add_acl_rule(self, policy_name: str, rule_config: Dict[str, Any]) -> bool:
        """
        Add a new rule to an existing ACL policy.
        
        Args:
            policy_name: Name of the ACL policy
            rule_config: Configuration for the new rule
            
        Returns:
            bool: True if rule added successfully, False otherwise
        """
        try:
            if policy_name not in self.active_policies:
                log.error(f"ACL policy '{policy_name}' not found")
                return False
            
            # This would require rebuilding the entire ACL with the new rule
            # For simplicity, we'll return not implemented for now
            log.warning("Adding individual ACL rules not yet implemented")
            return False
            
        except Exception as e:
            log.error(f"Exception adding ACL rule: {e}")
            return False
    
    def remove_acl_rule(self, policy_name: str, rule_index: int) -> bool:
        """
        Remove a rule from an existing ACL policy.
        
        Args:
            policy_name: Name of the ACL policy
            rule_index: Index of the rule to remove
            
        Returns:
            bool: True if rule removed successfully, False otherwise
        """
        try:
            if policy_name not in self.active_policies:
                log.error(f"ACL policy '{policy_name}' not found")
                return False
            
            # This would require rebuilding the entire ACL without the rule
            # For simplicity, we'll return not implemented for now
            log.warning("Removing individual ACL rules not yet implemented")
            return False
            
        except Exception as e:
            log.error(f"Exception removing ACL rule: {e}")
            return False
    
    # Implementation of SecurityManager abstract methods
    
    def create_policy(self, policy_config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Create an ACL policy from configuration.
        
        Args:
            policy_config: ACL policy configuration
            
        Returns:
            dict: Creation result
        """
        try:
            # Extract configuration parameters
            interface_config = policy_config.get('interface')
            deny_network = policy_config.get('deny_network')
            additional_rules = policy_config.get('additional_rules', [])
            
            if not interface_config:
                return create_result_dict(
                    success=False, 
                    error="Missing required interface configuration"
                )
            
            if not deny_network:
                return create_result_dict(
                    success=False, 
                    error="Missing required deny_network parameter"
                )
            
            # Convert to Interface object
            interface = self._dict_to_interface(interface_config)
            
            # Configure ACL
            success = self.configure_acl(
                ingress_interface=interface,
                deny_network=deny_network,
                additional_rules=additional_rules
            )
            
            policy_name = f"acl-{interface.interface_name}"
            
            if success:
                return create_result_dict(
                    success=True,
                    data={'policy_name': policy_name, 'status': 'active'}
                )
            else:
                return create_result_dict(success=False, error="Failed to configure ACL")
                
        except Exception as e:
            log.error(f"Exception creating ACL policy: {e}")
            return create_result_dict(success=False, error=str(e))
    
    def apply_policy(self, policy_name: str) -> bool:
        """
        Apply an ACL policy by name.
        
        Args:
            policy_name: Name of policy to apply
            
        Returns:
            bool: True if applied successfully, False otherwise
        """
        if policy_name not in self.active_policies:
            log.error(f"ACL policy '{policy_name}' not found")
            return False
        
        # Policy is already applied when created in this implementation
        log.info(f"ACL policy '{policy_name}' is already active")
        return True
    
    def remove_policy(self, policy_name: str) -> bool:
        """
        Remove an ACL policy by name.
        
        Args:
            policy_name: Name of policy to remove
            
        Returns:
            bool: True if removed successfully, False otherwise
        """
        if policy_name not in self.active_policies:
            log.warning(f"ACL policy '{policy_name}' not found")
            return True
        
        try:
            policy = self.active_policies[policy_name]
            
            # Create Interface object from stored data
            interface = Interface(interface_name=policy['interface'])
            
            # Remove the ACL configuration
            success = self.remove_acl_configuration(interface)
            
            return success
            
        except Exception as e:
            log.error(f"Exception removing ACL policy '{policy_name}': {e}")
            return False
    
    # Implementation of NetworkComponent abstract methods
    
    def create(self, config: Dict[str, Any]) -> Dict[str, Any]:
        """Create ACL configuration from config."""
        return self.create_policy(config)
    
    def delete(self, identifier: str) -> bool:
        """Delete ACL policy by name."""
        return self.remove_policy(identifier)
    
    def get_status(self, identifier: str) -> Dict[str, Any]:
        """Get status of ACL policy."""
        if identifier in self.active_policies:
            policy = self.active_policies[identifier]
            return {
                'name': identifier,
                'status': 'active',
                'interface': policy['interface'],
                'deny_network': policy['deny_network'],
                'rule_count': policy['rule_count'],
                'created_at': policy['created_at']
            }
        else:
            return {'name': identifier, 'status': 'not_found'}
    
    def list_components(self) -> List[Dict[str, Any]]:
        """List all active ACL policies."""
        return [
            {
                'name': name,
                'status': 'active',
                'interface': policy['interface'],
                'deny_network': policy['deny_network'],
                'rule_count': policy['rule_count']
            }
            for name, policy in self.active_policies.items()
        ]
    
    # Public utility methods
    
    def get_acl_statistics(self) -> Dict[str, Any]:
        """
        Get ACL statistics and hit counts.
        
        Returns:
            dict: ACL statistics
        """
        try:
            # This would implement actual ACL statistics retrieval from VPP
            # For now, return basic information
            return {
                'active_policies': len(self.active_policies),
                'policies': list(self.active_policies.keys()),
                'total_rules': sum(
                    policy['rule_count'] for policy in self.active_policies.values()
                ),
                'interfaces_protected': len(set(
                    policy['interface'] for policy in self.active_policies.values()
                ))
            }
            
        except Exception as e:
            log.error(f"Error getting ACL statistics: {e}")
            return {'error': str(e)}
    
    def verify_acl_configuration(self, policy_name: str) -> Dict[str, Any]:
        """
        Verify that an ACL policy is correctly configured in VPP.
        
        Args:
            policy_name: Name of the policy to verify
            
        Returns:
            dict: Verification results
        """
        if policy_name not in self.active_policies:
            return {'verified': False, 'error': 'Policy not found'}
        
        try:
            policy = self.active_policies[policy_name]
            
            # Check if ACL is configured in VPP
            current_config = self.connection.client.get_configuration()
            
            acl_configs = [
                item.config
                for item in current_config.items
                if isinstance(item.config, models.ACLConfigurationItem)
                and item.config.ingress == [policy['interface']]
                and item.config.name == policy_name
            ]
            
            acl_exists = len(acl_configs) > 0
            rule_count_matches = False
            
            if acl_exists:
                actual_rule_count = len(acl_configs[0].rules) if acl_configs[0].rules else 0
                rule_count_matches = actual_rule_count == policy['rule_count']
            
            return {
                'verified': acl_exists and rule_count_matches,
                'acl_exists': acl_exists,
                'rule_count_matches': rule_count_matches,
                'expected_rules': policy['rule_count'],
                'actual_rules': len(acl_configs[0].rules) if acl_exists and acl_configs[0].rules else 0
            }
            
        except Exception as e:
            log.error(f"Error verifying ACL configuration for '{policy_name}': {e}")
            return {'verified': False, 'error': str(e)}
    
    def create_advanced_acl_policy(self, policy_name: str, interface_name: str, 
                                  rules: List[Dict[str, Any]]) -> bool:
        """
        Create an advanced ACL policy with custom rules.
        
        Args:
            policy_name: Name for the new policy
            interface_name: Interface to apply ACL to
            rules: List of rule configurations
            
        Returns:
            bool: True if policy created successfully, False otherwise
        """
        try:
            log.info(f"Creating advanced ACL policy '{policy_name}' on {interface_name}")
            
            # Convert rule configurations to ACL rule objects
            acl_rules = []
            for rule_config in rules:
                rule = self._create_acl_rule_from_config(rule_config)
                if rule:
                    acl_rules.append(rule)
            
            # Create ACL configuration
            acl_config = models.ACLConfigurationItem(
                name=policy_name,
                ingress=[interface_name],
                rules=acl_rules,
            )
            
            # Apply configuration to VPP
            reply = self.connection.client.add_configuration(acl_config)
            
            errors = self._extract_reply_errors(reply)
            if errors:
                log.error(f"Error adding advanced ACL configuration: {'; '.join(errors)}")
                return False
            
            # Store policy information
            self.active_policies[policy_name] = {
                'interface': interface_name,
                'custom_rules': rules,
                'rule_count': len(acl_rules),
                'type': 'advanced',
                'created_at': self._get_current_timestamp()
            }
            
            log.info(f"Advanced ACL policy '{policy_name}' created successfully")
            return True
            
        except Exception as e:
            log.error(f"Exception creating advanced ACL policy: {e}")
            return False
    
    # Private helper methods
    
    def _create_acl_rule_from_config(self, rule_config: Dict[str, Any]) -> Optional[models.ACLRuleConfigurationItem]:
        """
        Create an ACL rule from configuration dictionary.
        
        Args:
            rule_config: Rule configuration
            
        Returns:
            Optional[ACLRuleConfigurationItem]: ACL rule object or None if invalid
        """
        try:
            action_str = rule_config.get('action', 'permit').upper()
            action = models.ACLAction.PERMIT if action_str == 'PERMIT' else models.ACLAction.DENY
            
            # Basic rule without refinement
            if not rule_config.get('refinement'):
                return models.ACLRuleConfigurationItem(action=action)
            
            refinement_config = rule_config['refinement']
            
            # Create IP specification if provided
            ip_spec = None
            if refinement_config.get('ip'):
                ip_config = refinement_config['ip']
                addresses = models.IPAddresses()
                
                if ip_config.get('source_network'):
                    addresses.source_network = ipaddress.IPv4Network(ip_config['source_network'])
                
                if ip_config.get('destination_network'):
                    addresses.destination_network = ipaddress.IPv4Network(ip_config['destination_network'])
                
                ip_spec = models.IPSpecification(addresses=addresses)
            
            return models.ACLRuleConfigurationItem(
                action=action,
                refinement=ip_spec
            )
            
        except Exception as e:
            log.error(f"Error creating ACL rule from config: {e}")
            return None
    
    def _dict_to_interface(self, interface_config: Dict[str, Any]) -> Interface:
        """Convert dictionary to Interface object."""
        return Interface(
            interface_name=interface_config.get('interface_name'),
            ip_address=interface_config.get('ip_address'),
            prefix_len=interface_config.get('prefix_len')
        )
    
    def _get_current_timestamp(self) -> str:
        """Get current timestamp as string."""
        from datetime import datetime
        return datetime.now().isoformat()
    
    def _find_acl_configs_in_vpp(self, current_config, interface_name: str) -> List:
        """Find ACL configurations for a specific interface in VPP config."""
        return [
            item.config
            for item in current_config.items
            if isinstance(item.config, models.ACLConfigurationItem)
            and interface_name in item.config.ingress
        ]
